// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.
//
// Original author: Johan Hoffman. Added to Kraftwerk 2002-01

#include "Discretiser.hh"
#include "FiniteElement.hh"
#include "ShapeFunction.hh"
#include "Equation.hh"
#include "EquationSystem.hh"
#include <Settings.hh>
#include <Globals.hh>

extern Settings *settings;
extern Globals *globals;

//-----------------------------------------------------------------------------
Discretiser::Discretiser(Grid *grid, Equation *equation)
{
  this->grid = grid;
  this->equation = equation;

  no_eq    = equation->noEq();
  no_nodes = grid->GetNoNodes();
  size     = no_eq * no_nodes;

  bc_function = globals->bc_function;
  if ( !bc_function )
	 display->Error("Boundary conditions are not specified.");
}
//-----------------------------------------------------------------------------
Discretiser::~Discretiser()
{
  
}
//-----------------------------------------------------------------------------
void Discretiser::Assemble(SparseMatrix *A, Vector *b)
{
  AssembleLHS(A);
  AssembleRHS(b);
  SetBoundaryConditions(A,b);
}
//-----------------------------------------------------------------------------
void Discretiser::AssembleLHS(SparseMatrix *A)
{
  int i,j;

  // Reallocate the matrix if necessary
  Allocate(A);
  
  // Reset the matrix
  A->Reset();
  
  real volume;
  real integral;

  FiniteElement element_trial(grid,no_eq);
  FiniteElement element_test(grid,no_eq);

  ShapeFunction *u;
  ShapeFunction *v;

  int no_dof;

  // Go through all elements and assemble
  for (int el=0;el<grid->GetNoCells();el++){

    // Compute element data
    element_trial.Refill(el);
    element_test.Refill(el);
    
    // Get the number of dof
    no_dof = element_trial.GetDim();
    
    volume = element_trial.GetVolume();
    
    // Compute element matrix and put the entries directly into A
    for (int test=0;test<no_dof;test++){
      for (int trial=0;trial<no_dof;trial++){
	for (int k=0;k<no_eq;k++){
	  for (int l=0;l<no_eq;l++){
				
	    u = element_trial.GetShapeFunction(trial,l);
	    v = element_test.GetShapeFunction(test,k);
				
		 if ( no_eq == 1 )
			integral = equation->IntegrateLHS(u[0],v[0]);
		 else
			integral = ((EquationSystem *) equation)->IntegrateLHS(u,v);
				
	    i = element_test.GetGlobalDof(test)*no_eq + k;
	    j = element_trial.GetGlobalDof(trial)*no_eq + l;
				
	    A->Add(i,j,integral*volume);
	    
	  }
	}
      }
    }
  }
  
  // Drop zero elements from matrix
  //A->DropZero(1e-12);
}
//-----------------------------------------------------------------------------
void Discretiser::AssembleRHS(Vector *b)
{
  Cell *c;
  int i,j;

  // Allocate the vector if we have to
  Allocate(b);
  
  // Reset the matrix
  b->SetToConstant(0.0);
  
  real volume;
  real integral;
  
  FiniteElement element_test(grid,no_eq);

  ShapeFunction *v;

  int no_dof;

  // Go through all elements and assemble
  for (int el=0;el<grid->GetNoCells();el++){
    
    // Compute element data
    element_test.Refill(el);
    
    // Get the number of dof
    no_dof = element_test.GetDim();
    
    volume = element_test.GetVolume();
    
    for (int test=0;test<no_dof;test++){
      for (int k=0;k<no_eq;k++){
	
	v = element_test.GetShapeFunction(test,k);

		 if ( no_eq == 1 )
			integral = equation->IntegrateRHS(v[0]);
		 else
			integral = ((EquationSystem *) equation)->IntegrateRHS(v);
	
		 i = element_test.GetGlobalDof(test)*no_eq + k;
		 b->Set(i,integral*volume);
	
      }
    }
  }
  
  // Drop zero elements from matrix
  //A->DropZero(1e-12);
}
//-----------------------------------------------------------------------------
void Discretiser::SetBoundaryConditions(SparseMatrix *A, Vector *b)
{
  if ( (A->Size(0) != no_eq*no_nodes ) || ( b->Size() != no_eq*no_nodes ) )
  	 display->Error("You must assemble the matrix before settings boundary conditions.");

  Point p;
  kw_bc bc;
  
  for (int i=0;i<no_nodes;i++){
    for (int component=0;component<no_eq;component++){
		
      p  = grid->GetNode(i)->GetCoord();
      bc = bc_function(p.x,p.y,p.z,component);
      
      switch ( bc.type ){
      case dirichlet:
		  A->SetRowIdentity(i*no_eq+component);
		  b->Set(i*no_eq+component,bc.val);
		  break;
      case neumann:
		  if ( bc.val != 0.0 )
			 display->Error("Inhomogeneous Neumann boundary conditions not implemented.");
		  break;
      default:
		  display->InternalError("Discretiser::SetBoundaryConditions()",
										 "Unknown boundary condition type.");
      }
    } 
  }
  
}
//-----------------------------------------------------------------------------
void Discretiser::Allocate(SparseMatrix *A)
{
  int ncols[size];
  int ii;
  bool ok = true;
  
  // Compute the number of columns for every row in the matrix
  for(int i=0;i<no_nodes;i++)
	 for (int j=0;j<no_eq;j++){
		ii = i*no_eq+j;
		ncols[ii] = no_eq * grid->GetNode(i)->GetNoNodeNeighbors();

		if ( ncols[ii] != A->GetRowLength(ii) )
		  ok = false;
		
	 }
  
  // Reallocate the matrix with the new size
  if ( !ok )
	 A->Resize(size,size,ncols);
}
//-----------------------------------------------------------------------------
void Discretiser::Allocate(Vector *b)
{
  if ( b->Size() != size )
	 b->Resize(size);
}
//-----------------------------------------------------------------------------
