// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "Field.hh"
#include <Display.hh>
#include <Settings.hh>
//#include <ShapeFunctionTetLin_old.hh>
//#include <ShapeFunctionTriLin_old.hh>


extern Settings *settings;

//-----------------------------------------------------------------------------
Field::Field()
{

  pi = 3.141592653589793238462;

  rep = NONE;
  whatFcn = UNDEF;

  grid_init = 0;

  quadorder = 1;

  UNIFORM_NODAL_BASIS = false;
};
//-----------------------------------------------------------------------------
Field::Field( Grid* g )
{

  pi = 3.141592653589793238462;

  rep = NONE;
  whatFcn = UNDEF;
  
  grid = g;
  grid_init = 1;

  noNodes    = grid->GetNoNodes();
  noElms     = grid->GetNoCells();

  settings->Get("dimension",&noSpaceDim);

  coord.newsize(noSpaceDim);
  locgrad.newsize(noSpaceDim);

  quadorder = 1;

  UNIFORM_NODAL_BASIS = false;
  if ( g->AllCellsEqual() ){
    UNIFORM_NODAL_BASIS = true;
    GetNodalBasis(0);
    switch( locBasis ){ 
    case TRI_BASIS: 
      locnodes.newsize(3);
      quadrule.init(TRI,quadorder);
      quadrule.getQuadratureRule( quadpoints, quadweights );
      noQuadPoints = quadweights.size();
      break;
    case TET_BASIS: 
      locnodes.newsize(4);
      quadrule.init(TET,quadorder);
      quadrule.getQuadratureRule( quadpoints, quadweights );
      noQuadPoints = quadweights.size();
      break;
    default:
      cout << "ERROR: The nodal basis is not implemented" << endl;
      exit(1);
    }
  }

}
//-----------------------------------------------------------------------------
int Field::AttachGrid( Grid* g )
{

  grid = g;
  grid_init = 1;

  noNodes = grid->GetNoNodes();
  noElms  = grid->GetNoCells();

  settings->Get("dimension",&noSpaceDim);
  
  coord.newsize(noSpaceDim);
  locgrad.newsize(noSpaceDim);  

  UNIFORM_NODAL_BASIS = false;
  if ( (*g).AllCellsEqual() ){
    UNIFORM_NODAL_BASIS = true;
    GetNodalBasis(0);
    switch( locBasis ){ 
    case TRI_BASIS: 
      locnodes.newsize(3);
      quadrule.init(TRI,quadorder);
      quadrule.getQuadratureRule( quadpoints, quadweights );
      noQuadPoints = quadweights.size();
      break;
    case TET_BASIS: 
      locnodes.newsize(4);
      quadrule.init(TET,quadorder);
      quadrule.getQuadratureRule( quadpoints, quadweights );
      noQuadPoints = quadweights.size();
      break;
    default:
      cout << "ERROR: The nodal basis is not implemented" << endl;
      exit(1);
    }
  }

  return 0;
}
//-----------------------------------------------------------------------------
Grid* Field::GetGrid()
{
  return grid;
}
//----------------------------------------------------------------------------- 
void Field::operator = ( Field* fld )
{
  pi = 3.141592653589793238462;

  rep = fld->GetRep();
  whatFcn = fld->GetFcn();
  
  if ( fld->grid_init == 1 ) AttachGrid( fld->GetGrid() );
} 
//-----------------------------------------------------------------------------
int Field::GetNodalBasis( int el )
{
  noBasisFcns = grid->GetCell(el)->GetSize();
  
  if ( noSpaceDim == 1 ){
    cout << "ERROR: Not implemented for 1d" << endl;
    exit(1);
  }
  if ( noSpaceDim == 2 ){
    switch( noBasisFcns ){ 
    case 3: 
      locBasis = TRI_BASIS;
      break;
    case 4:
      locBasis = QUAD_BASIS;
      break;
    default:
      cout << "ERROR: The nodal basis is not implemented" << endl;
      exit(1);
    }
  }  
  if ( noSpaceDim == 3 ){
    switch( noBasisFcns ){ 
    case 4: 
      locBasis = TET_BASIS;
      break;
    case 8:
      locBasis = HEX_BASIS;
      break;
    default:
      cout << "ERROR: The nodal basis is not implemented" << endl;
      exit(1);
    }
  }  

  return 0;
}
//-----------------------------------------------------------------------------
int Field::GetElementVector( MV_Vector<real>& elVec )
{

  elVec.newsize(noElms);

  int el,nod,dim,noNodInEl;
  switch( rep ){ 
  case NODAL_VECTOR: 
    for ( el=0; el < noElms; el++ ) elVec(el) = Nodes2elms(el);
    break;
  case NODAL_MATRIX: 
    for ( el=0; el < noElms; el++ ) elVec(el) = Nodes2elms(el);
    break;
  case ELEMENT_VECTOR:
    for ( el=0; el < noElms; el++ ) elVec(el) = (*elementValuesVector)(el);
    break;
  case ELEMENT_MATRIX:
    for ( el=0; el < noElms; el++ ) elVec(el) = (*elementValuesMatrix)(el,component);
    break;
  case CONSTANT:
    for ( el=0; el < noElms; el++ ) elVec(el) = (*constant);
    break;
  case FUNCTION:
    for ( el=0; el < noElms; el++ ) elVec(el) = Eval(el);
    break;
  case NONE:
    printf("ERROR: Field is not initialized!!\n");
    exit(1);
  default:
    printf("ERROR: Wrong Field representation!!\n");
    exit(1);
  }
  
  return 0;
}
//-----------------------------------------------------------------------------
void Field::GetLocalNodes(int el, real *local_nodes)
{

  switch( rep ){ 
  case NODAL_VECTOR: 
    for (int i=0; i < noBasisFcns; i++) local_nodes[i] = (*nodalValuesVector)(grid->GetCell(el)->GetNode(i)); 
    break;
  case NODAL_MATRIX: 
    for (int i=0; i < noBasisFcns; i++) local_nodes[i] = (*nodalValuesMatrix)(grid->GetCell(el)->GetNode(i),component); 
    break;
  case ELEMENT_VECTOR:
    for (int i=0; i < noBasisFcns; i++) local_nodes[i] = (*elementValuesVector)(el); 
    break;
  case ELEMENT_MATRIX:
    for (int i=0; i < noBasisFcns; i++) local_nodes[i] = (*elementValuesMatrix)(el,component); 
    break;
  case CONSTANT:
    for (int i=0; i < noBasisFcns; i++) local_nodes[i] = (*constant);
    break;
  case FUNCTION:
    Node *n;
    for (int i=0; i < noBasisFcns; i++ ){
      n = grid->GetNode(grid->GetCell(el)->GetNode(i));
      for (int sd=0; sd < noSpaceDim; sd++) coord(sd) = n->GetCoord(sd);
      local_nodes[i] = Eval(coord);
    }
    break;
  case NONE:
    printf("ERROR: Field is not initialized!!\n");
    exit(1);
  default:
    printf("ERROR: Wrong Field representation!!\n");
    exit(1);
  }  
}
//-----------------------------------------------------------------------------
int Field::GetNodalVector( MV_Vector<real>& nodVec )
{

  nodVec.newsize(noNodes);

  int el,nod,sd;
  switch( rep ){ 
  case NODAL_VECTOR: 
    for ( nod=0; nod < noNodes; nod++ ) nodVec(nod) = (*nodalValuesVector)(nod);
    break;
  case NODAL_MATRIX: 
    for ( nod=0; nod < noNodes; nod++ ) nodVec(nod) = (*nodalValuesMatrix)(nod,component);
    break;
  case ELEMENT_VECTOR:
    for ( nod=0; nod < noNodes; nod++ ) nodVec(nod) = Elms2nodes(nod);
    break;
  case ELEMENT_MATRIX:
    for ( nod=0; nod < noNodes; nod++ ) nodVec(nod) = Elms2nodes(nod);
    break;
  case CONSTANT:
    for ( nod=0; nod < noNodes; nod++ ) nodVec(nod) = (*constant);
    break;
  case FUNCTION:
	 Node *n;
    for ( nod=0; nod < noNodes; nod++ ){
		n = grid->GetNode(nod);
      for ( sd=0; sd < noSpaceDim; sd++ )
		  coord(sd) = n->GetCoord(sd);
      nodVec(nod) = Eval(coord);
    }
    break;
  case NONE:
    printf("ERROR: Field is not initialized!!\n");
    exit(1);
  default:
    printf("ERROR: Wrong Field representation!!\n");
    exit(1);
  }
  
  return 0;
}
//-----------------------------------------------------------------------------
real Field::Nodes2elms( int el )
{

  int noNodesInElm;
  int i;
  
  real elVal;
  if ( rep == NODAL_VECTOR ){
    elVal = 0.0;
    noNodesInElm = grid->GetCell(el)->GetSize();
    for ( i=0; i < noNodesInElm; i++ ) elVal += (*nodalValuesVector)(grid->GetCell(el)->GetNode(i));
    elVal /= noNodesInElm;
  } else if ( rep == NODAL_MATRIX ){
    elVal = 0.0;
    noNodesInElm = grid->GetCell(el)->GetSize();
    for ( i=0; i < noNodesInElm; i++ ) elVal += (*nodalValuesMatrix)(grid->GetCell(el)->GetNode(i),component);
    elVal /= noNodesInElm;
  }
  else{
    cout << "Field must have a nodal representation!" << endl;
    exit(1);
  }
  
  return elVal;
}  
//-----------------------------------------------------------------------------
real Field::Elms2nodes( int node )
{

  cout << "WARNING: This function is memory consuming" << endl;

  int i;
  real nodVal;

  Node *n;
  n = grid->GetNode(node);

  if ( rep == ELEMENT_VECTOR ){
    nodVal = 0.0;
    for ( i=0; i < n->GetNoCellNeighbors(); i++ ){
      nodVal += (*elementValuesVector)(n->GetCellNeighbor(i));
    }
    nodVal /= double(  n->GetNoCellNeighbors() );
  }
  else if ( rep == ELEMENT_MATRIX ){
    nodVal = 0.0;
    for ( i=0; i < n->GetNoCellNeighbors(); i++ ){
      nodVal += (*elementValuesMatrix)(n->GetCellNeighbor(i),component);
    }
    nodVal /= double( n->GetNoCellNeighbors() );
  }
  else{
    cout << "Field must have an element representation!" << endl;
    exit(1);
  }

 return nodVal;
}  
//-----------------------------------------------------------------------------
int Field::InitByNodes( MV_Vector<real>* nodVal )
{

  nodalValuesVector = nodVal;
  rep = NODAL_VECTOR;

  return 0;
}
//-----------------------------------------------------------------------------    
int Field::InitByNodes( MV_ColMat<real>* nodVal, int col )
{

  nodalValuesMatrix = nodVal;
  component = col;
  rep = NODAL_MATRIX;

  return 0;
}
//-----------------------------------------------------------------------------    
int Field::InitByElms( MV_Vector<real>* elVal )
{

  elementValuesVector = elVal;
  rep = ELEMENT_VECTOR;

  return 0;
}
//-----------------------------------------------------------------------------    
int Field::InitByElms( MV_ColMat<real>* elVal, int col )
{

  elementValuesMatrix = elVal;
  component = col;
  rep = ELEMENT_MATRIX;

  return 0;
}
//-----------------------------------------------------------------------------    
int Field::InitByConstant( real* c )
{

  constant = c;
  rep = CONSTANT;

  return 0;
}
//-----------------------------------------------------------------------------    
int Field::InitByFcn( MV_Vector<real> par, FunctionClass fcntype )
{
  
  parameters.newsize(par.size());
  for ( int i=0; i < par.size(); i++ ) parameters(i) = par(i);
  rep = FUNCTION;
  whatFcn = fcntype;

  return 0;
}
//-----------------------------------------------------------------------------
real Field::Eval()
{

  real value;

  switch( rep ){ 
  case NODAL_VECTOR: 
    cout << "ERROR: Must specify a point for evaluation!!" << endl;
    exit(1);
    break;
  case NODAL_MATRIX: 
    cout << "ERROR: Must specify a point for evaluation!!" << endl;
    exit(1);
    break;
  case ELEMENT_VECTOR:
    cout << "ERROR: Must specify a point or element for evaluation!!" << endl;
    exit(1);
    break;
  case ELEMENT_MATRIX:
    cout << "ERROR: Must specify a point or element for evaluation!!" << endl;
    exit(1);
    break;
  case CONSTANT:
    value  = (*constant);
    break;
  case FUNCTION:
    cout << "ERROR: Must specify a point for evaluation!!" << endl;
    exit(1);
    break;
  case NONE:
    printf("ERROR: Field is not initialized!!\n");
    exit(1);
  default:
    printf("ERROR: Wrong Field representation!!\n");
    exit(1);
  }

  return value;
}
//-----------------------------------------------------------------------------
real Field::Eval( MV_Vector<real>& point )
{

  real value;
  
  coord = point;
  switch( rep ){ 
  case NODAL_VECTOR: 
    EvalNodalField( value );
    break;
  case NODAL_MATRIX: 
    EvalNodalField( value );
    break;
  case ELEMENT_VECTOR:
    EvalElementField( value );
    break;
  case ELEMENT_MATRIX:
    EvalElementField( value );
    break;
  case CONSTANT:
    value  = (*constant);
    break;
  case FUNCTION:
    EvalFunctionField( value );
    break;
  case NONE:
    cout << "ERROR: Field is not initialized" << endl;
    exit(1);
  default:
    cout << "ERROR: Wrong Field representation" << endl;
    exit(1);
  }

  return value;
  
}
//----------------------------------------------------------------------------- 
real Field::Eval( MV_Vector<real>& point, int el )
{

  real value;

  coord = point;
  switch( rep ){ 
  case NODAL_VECTOR: 
    EvalNodalField( value, el );
    break;
  case NODAL_MATRIX: 
    EvalNodalField( value, el );
    break;
  case ELEMENT_VECTOR:
    value  = (*elementValuesVector)(el);
    break;
  case ELEMENT_MATRIX:
    value  = (*elementValuesMatrix)(el,component);
    break;
  case CONSTANT:
    value  = (*constant);
    break;
  case FUNCTION:
    EvalFunctionField( value );
    break;
  case NONE:
    cout << "ERROR: Field is not initialized" << endl;
    exit(1);
  default:
    cout << "ERROR: Wrong Field representation" << endl;
    exit(1);
  }

  return value;
  
}
//-----------------------------------------------------------------------------
real Field::Eval( int el )
{
  real value;
  Node *n;
  Cell *c;
  
  switch( rep ){ 
  case NODAL_VECTOR: 
    value = Nodes2elms(el);
    break;
  case NODAL_MATRIX: 
    value = Nodes2elms(el);
    break;
  case ELEMENT_VECTOR:
    value  = (*elementValuesVector)(el);
    break;
  case ELEMENT_MATRIX:
    value  = (*elementValuesMatrix)(el,component);
    break;
  case CONSTANT:
    value  = (*constant);
    break;
  case FUNCTION:
	 c = grid->GetCell(el);
    for (int i=0;i<noSpaceDim;i++){
      coord(i) = 0.0;
      for (int j=0;j<c->GetSize();j++){
		  n = grid->GetNode(c->GetNode(j));
		  coord(i) += n->GetCoord(i);
		}
      coord(i) /= double(c->GetSize());
    }
    EvalFunctionField( value );
    break;
  case NONE:
    printf("ERROR: Field is not initialized!!\n");
    exit(1);
  default:
    printf("ERROR: Wrong Field representation!!\n");
    exit(1);
  }

  return value;
  
}
//-----------------------------------------------------------------------------
int Field::EvalNodalField( real& value )
{
  
  // This function computes the value in the point = coord, 
  // of a Field in nodal repsentation. 

  if ( (rep!=NODAL_VECTOR) && (rep!=NODAL_MATRIX) ){
    cout << "ERROR: Not a nodal representation" << endl;
    exit(1);
  }

  int i, el = -1;
  for ( i=0; i < noElms; i++ ){
    if ( CheckIfInElm(i) == 1 ){
      el = i;
      break;
    }
  }
  if ( el == -1 ){
    cout << "ERROR: Point is not in any element" << endl;
    exit(1);
  }

  value = 0.0;
  if ( rep == NODAL_VECTOR ){
    for ( i=0; i < noBasisFcns; i++ ) value += barycoord(i) * (*nodalValuesVector)(grid->GetCell(el)->GetNode(i)); 
  }
  if ( rep == NODAL_MATRIX ){
    for ( i=0; i < noBasisFcns; i++ ) value += barycoord(i) * (*nodalValuesMatrix)(grid->GetCell(el)->GetNode(i),component); 
  }

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalElementField( real& value )
{

  // This function computes the value in the point 
  // coor[], of a Field in element repsentation. 

  if ( (rep!=ELEMENT_VECTOR) && (rep!=ELEMENT_MATRIX) ){
    cout << "ERROR: Not an element representation" << endl;
    exit(1);
  }

  int i, el = -1;
  for ( i=0; i < noElms; i++ ){
    if ( CheckIfInElm(i) == 1 ){
      el  = i;
      break;
    }
  }
  if ( el == -1 ){
    cout << "ERROR: Point is not in any element" << endl;
    exit(1);
  }

  if ( rep == ELEMENT_VECTOR ) value = (*elementValuesVector)(el);
  if ( rep == ELEMENT_MATRIX ) value = (*elementValuesMatrix)(el,component);

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalConstantField( real& value )
{

  if ( rep != CONSTANT ){
    cout << "ERROR: Not a constant representation" << endl;
    exit(1);
  }
  value = (*constant);

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalNodalField( real& value, int el )
{
  
  // This function computes the value in the point 
  // coor[], of a Field in nodal repsentation. 

  if ( (rep!=NODAL_VECTOR) && (rep!=NODAL_MATRIX) ){
    cout << "ERROR: Not a nodal representation" << endl;
    exit(1);
  }

  GetBarycentric( el ); 

  value = 0.0;

  int i;
  if ( rep == NODAL_VECTOR ){
    for ( i=0; i < noBasisFcns; i++ ) value += barycoord(i) * (*nodalValuesVector)(grid->GetCell(el)->GetNode(i)); 
  }
  if ( rep == NODAL_MATRIX ){
    for ( i=0; i < noBasisFcns; i++ ) value += barycoord(i) * (*nodalValuesMatrix)(grid->GetCell(el)->GetNode(i),component); 
  }

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalElementField( real& value, int el )
{

  if ( (rep!=ELEMENT_VECTOR) && (rep!=ELEMENT_MATRIX) ){
    cout << "ERROR: Not an element representation" << endl;
    exit(1);
  }

  if ( rep == ELEMENT_VECTOR ) value = (*elementValuesVector)(el);
  if ( rep == ELEMENT_MATRIX ) value = (*elementValuesMatrix)(el,component);

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalConstantField( real& value, int el )
{
  
  if ( rep != CONSTANT ){
    cout << "ERROR: Not a constant representation" << endl;
    exit(1);
  }

  value = (*constant);

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalFunctionField( real& value )
{

  if ( rep != FUNCTION ){
    cout << "ERROR: Not a function representation" << endl;
    exit(1);
  }

  switch( whatFcn ){ 
  case POLYNOMIAL: 
    EvalPolynomialField( value );
    break;
  case HOLDERFCN:
    EvalHolderField( value );
    break;
  case HOLDERFCNTD:
    EvalTDHolderField( value );
    break;
  case HOLDERFCNROT:
    EvalHolderFieldRot( value );
    break;
  case WEIERSTRASSFCNTD:
    EvalTDWeierstrassField( value );
    break;
  case WEIERSTRASSROT:
    EvalWeierstrassFieldRot( value );
    break;
  case WEIERSTRASSTDVELX:
    EvalTDVelXWeierstrassField( value );
    break;
  case WEIERSTRASSTDVELY:
    EvalTDVelYWeierstrassField( value );
    break;
  case HOLDERFCNTDVELX:
    EvalTDVelXHolderField( value );
    break;
  case HOLDERFCNTDVELY:
    EvalTDVelYHolderField( value );
    break;
  case HOLDERFCN2D:
    Eval2DHolderField( value );
    break;
  case WEIERSTRASS:
    EvalWeierstrassField( value );
    break;
  case NONE:
    printf("ERROR: Field is not initialized!!\n");
    exit(1);
  default:
    printf("ERROR: Wrong Field representation!!\n");
    exit(1);
  }
  
  return 0;
}
//-----------------------------------------------------------------------------
real Field::GetNodalValue( int node )
{
  switch( rep ){ 
  case NODAL_VECTOR: 
    return (*nodalValuesVector)(node);
    break;
  case NODAL_MATRIX: 
    return (*nodalValuesMatrix)(node,component);
    break;
  case ELEMENT_VECTOR:
    return Elms2nodes(node);
    break;
  case ELEMENT_MATRIX:
    return Elms2nodes(node);
    break;
  case CONSTANT:
    return (*constant);
    break;
  case FUNCTION:
    cout << "ERROR: Not nodal representation!!" << endl;
    exit(1);
    break;
  case NONE:
    printf("ERROR: Field is not initialized!!\n");
    exit(1);
  default:
    printf("ERROR: Wrong Field representation!!\n");
    exit(1);
  }

  return (*nodalValuesVector)(node);
}
//-----------------------------------------------------------------------------
real Field::GetElemValue( int el )
{
  
  switch( rep ){ 
  case NODAL_VECTOR: 
    return Nodes2elms(el);
    break;
  case NODAL_MATRIX: 
    return Nodes2elms(el);
    break;
  case ELEMENT_VECTOR:
    return (*elementValuesVector)(el);
    break;
  case ELEMENT_MATRIX:
    return (*elementValuesMatrix)(el,component);
    break;
  case CONSTANT:
    return (*constant);
    break;
  case FUNCTION:
    cout << "ERROR: Not nodal representation!!" << endl;
    exit(1);
    break;
  case NONE:
    cout << "ERROR: Field is not initialized" << endl;
    exit(1);
  default:
    cout << "ERROR: Field is not initialized" << endl;
    exit(1);
  }

  return (*elementValuesVector)(el);
}
//-----------------------------------------------------------------------------
real Field::EvalGradient( int el, int comp )
{

  real value;

  CompGradient(el);
  value = locgrad(comp);

  return value;
}
//-----------------------------------------------------------------------------
int Field::EvalGradient( int el, MV_Vector<real>& grad )
{

  CompGradient(el);
  grad = locgrad;

  return 0;
}
//-----------------------------------------------------------------------------
int Field::CompGradient( int el )
{

  if ( !UNIFORM_NODAL_BASIS ) GetNodalBasis(el);
  switch( locBasis ){ 
  case TRI_BASIS: 
    CompGradientTri(el);
    break;
  case QUAD_BASIS: 
    CompGradientQuad(el);
    break;
  case TET_BASIS: 
    CompGradientTet(el);
    break;
  default:
    cout << "ERROR: The nodal basis is not implemented" << endl;
    exit(1);
  }

  return 0;
}
//-----------------------------------------------------------------------------
int Field::CompGradientQuad( int el )
{

  // This function computes the local gradient locgrad of a function in the 2D unit square, 
  // on a uniform mesh.                                                       

  display->InternalError("Field::GetBarycentricQuad()","Quads not implemented.");
  
 //  if ( (rep!=NODAL_VECTOR) && (rep!=NODAL_MATRIX) ){
//     cout << "ERROR: Not a nodal representation" << endl;
//     exit(1);
//   }

//   locgrad.newsize(noSpaceDim);
  
//   real h = 1/sqrt(noElms); 
  
//   switch( rep ){ 
//   case NODAL_VECTOR: 
//     locgrad(0) = ( (*nodalValuesVector)(grid->GetGlobalNode(el,1)) - (*nodalValuesVector)(grid->GetGlobalNode(el,0)) 
// 						 + (*nodalValuesVector)(grid->GetGlobalNode(el,2)) - (*nodalValuesVector)(grid->GetGlobalNode(el,3)) ) / (2*h);
//     locgrad(1) = ( (*nodalValuesVector)(grid->GetGlobalNode(el,3)) - (*nodalValuesVector)(grid->GetGlobalNode(el,0))
// 						 + (*nodalValuesVector)(grid->GetGlobalNode(el,2)) - (*nodalValuesVector)(grid->GetGlobalNode(el,1)) ) / (2*h);
//     break;
//   case NODAL_MATRIX: 
//     locgrad(0) = ( (*nodalValuesMatrix)(grid->GetGlobalNode(el,1),component) - (*nodalValuesMatrix)(grid->GetGlobalNode(el,0),component) 
// 						 + (*nodalValuesMatrix)(grid->GetGlobalNode(el,2),component) - (*nodalValuesMatrix)(grid->GetGlobalNode(el,3),component) ) / (2*h);
//     locgrad(1) = ( (*nodalValuesMatrix)(grid->GetGlobalNode(el,3),component) - (*nodalValuesMatrix)(grid->GetGlobalNode(el,0),component)
// 						 + (*nodalValuesMatrix)(grid->GetGlobalNode(el,2),component) - (*nodalValuesMatrix)(grid->GetGlobalNode(el,1),component) ) / (2*h);
//     break;
//   default:
//     cout << "ERROR: Not a nodal representation" << endl;
//     exit(1);
//   }
  
  return 0;
}
//-----------------------------------------------------------------------------
int Field::CompGradientTri( int el )
{

  // This function computes the local gradient locgrad of a function in a triangular mesh.  
  
  if ( (rep!=NODAL_VECTOR) && (rep!=NODAL_MATRIX) ){
    cout << "ERROR: Not a nodal representation" << endl;
    exit(1);
  }

  if ( !UNIFORM_NODAL_BASIS ) locnodes.newsize(grid->GetCell(el)->GetSize());

  Cell *c = grid->GetCell(el);
  
  for (int i=0;i<c->GetSize();i++)
	 locnodes(i) = c->GetNode(i);
  
  Node *n1 = grid->GetNode(locnodes(0));
  Node *n2 = grid->GetNode(locnodes(1));
  Node *n3 = grid->GetNode(locnodes(2));
  
  real x1 = n1->GetCoord(0);
  real x2 = n2->GetCoord(0);
  real x3 = n3->GetCoord(0);
  real y1 = n1->GetCoord(1);
  real y2 = n2->GetCoord(1);
  real y3 = n3->GetCoord(1);
  
  dN2x = y3-y1;  
  dN2y = x1-x3;  
  dN3x = y1-y2;  
  dN3y = x2-x1;

  real det = dN3y*dN2x - dN3x*dN2y;  
  real d = 1/det;
  
  dN2x *= d;  
  dN2y *= d;  
  dN3x *= d;  
  dN3y *= d;
  dN1x = - dN2x - dN3x;  
  dN1y = - dN2y - dN3y;
  
  switch( rep ){ 
  case NODAL_VECTOR: 
    locgrad(0)  = (*nodalValuesVector)(locnodes(0))*dN1x;
    locgrad(0) += (*nodalValuesVector)(locnodes(1))*dN2x;
    locgrad(0) += (*nodalValuesVector)(locnodes(2))*dN3x;
    locgrad(1)  = (*nodalValuesVector)(locnodes(0))*dN1y;
    locgrad(1) += (*nodalValuesVector)(locnodes(1))*dN2y;
    locgrad(1) += (*nodalValuesVector)(locnodes(2))*dN3y;
    break;
  case NODAL_MATRIX: 
    locgrad(0)  = (*nodalValuesMatrix)(locnodes(0),component)*dN1x;
    locgrad(0) += (*nodalValuesMatrix)(locnodes(1),component)*dN2x;
    locgrad(0) += (*nodalValuesMatrix)(locnodes(2),component)*dN3x;
    locgrad(1)  = (*nodalValuesMatrix)(locnodes(0),component)*dN1y;
    locgrad(1) += (*nodalValuesMatrix)(locnodes(1),component)*dN2y;
    locgrad(1) += (*nodalValuesMatrix)(locnodes(2),component)*dN3y;
    break;
  default:
    cout << "ERROR: Not a nodal representation" << endl;
    exit(1);
  }

  return 0;
}
//-----------------------------------------------------------------------------
int Field::CompGradientTet( int el )
{  
  /*
  // This function computes the local gradient locgrad of a function in a tetrahedral mesh.  
  
  if ( (rep!=NODAL_VECTOR) && (rep!=NODAL_MATRIX) ){
    cout << "ERROR: Not a nodal representation" << endl;
    exit(1);
  }
  
  if ( !UNIFORM_NODAL_BASIS ) locnodes.newsize(grid->GetCell(el)->GetSize());
  
  int i;
  for ( i=0; i < grid->GetCell(el)->GetSize(); i++ )
    locnodes(i) = grid->GetCell(el)->GetNode(i);
  
  real local_nodal_values[4];
  real local_gradient[3];
  
  switch( rep ){ 
  case NODAL_VECTOR: 
    local_nodal_values[0] = (*nodalValuesVector)(locnodes(0));
    local_nodal_values[1] = (*nodalValuesVector)(locnodes(1));
    local_nodal_values[2] = (*nodalValuesVector)(locnodes(2));
    local_nodal_values[3] = (*nodalValuesVector)(locnodes(3));
    break;
  case NODAL_MATRIX: 
    local_nodal_values[0] = (*nodalValuesMatrix)(locnodes(0),component);
    local_nodal_values[1] = (*nodalValuesMatrix)(locnodes(1),component);
    local_nodal_values[2] = (*nodalValuesMatrix)(locnodes(2),component);
    local_nodal_values[3] = (*nodalValuesMatrix)(locnodes(3),component);
    break;
  default:
    cout << "ERROR: Not a nodal representation" << endl;
    exit(1);
  }

  ShapeFunctionTetLin_old sfcn(grid);
  sfcn.Refill(el);
  sfcn.EvalGradient(local_nodal_values,local_gradient);
 
  for (int i=0; i < 3; i++ ) locgrad(i) = local_gradient[i];
  */
  return 0;
}
//-----------------------------------------------------------------------------
int Field::Integrate( real& integral )
{

  int el;
  integral = 0.0;
  for ( el=0; el < noElms; el++ ){
    if ( !UNIFORM_NODAL_BASIS ) GetNodalBasis(el);
    switch( locBasis ){ 
    case LINE_BASIS:
      //IntegrateQuadratureLine(integral,el);
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    case TRI_BASIS: 
      IntegrateQuadratureTri(integral,el);
      break;
    case QUAD_BASIS:
      //IntegrateQuadratureQuad(integral,el);
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    case TET_BASIS:
      IntegrateQuadratureTet(integral,el);
      break;
    case HEX_BASIS:
      //IntegrateQuadratureHex(integral,el);
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    case PYR_BASIS:
      //IntegrateQuadraturePyr(integral,el);
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    case PRISM_BASIS:
      //IntegrateQuadraturePrism(integral,el);
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    default:
      cout << "ERROR: Nodal basis not implemented" << endl;
      exit(1);
      break; 
    }
  }

  return 0;
}
//-----------------------------------------------------------------------------
int Field::IntegrateQuadratureTri( real& integral, int el )
{

  if ( !UNIFORM_NODAL_BASIS ){
    quadrule.init(TRI,quadorder);
    quadrule.getQuadratureRule( quadpoints, quadweights );
    noQuadPoints = quadweights.size();
  }

  determinant = 2.0 * grid->GetCell(el)->ComputeVolume(grid);

  int qp,sd,sf;
  Cell *c = grid->GetCell(el);
  Node *n;
  
  for ( qp=0; qp < noQuadPoints; qp++ ){
    for ( sd=0; sd < noSpaceDim; sd++ ){
      coord(sd) = 0.0;
      for ( sf=0; sf < grid->GetCell(el)->GetSize(); sf++ ){
		  n = grid->GetNode(c->GetNode(sf));
		  coord(sd) += quadpoints(qp,sd) * ( n->GetCoord(sd) );
      }
    }
    integral += Eval(coord,el) * quadweights(qp) * determinant;
  }

  return 0;
}
//-----------------------------------------------------------------------------
int Field::IntegrateQuadratureTet( real& integral, int el )
{

  if ( !UNIFORM_NODAL_BASIS ){
    quadrule.init(TET,quadorder);
    quadrule.getQuadratureRule( quadpoints, quadweights );
    noQuadPoints = quadweights.size();
  }

  determinant = 6.0 * grid->GetCell(el)->ComputeVolume(grid);

  Cell *c = grid->GetCell(el);
  Node *n;
  
  int qp,sd,sf;
  for ( qp=0; qp < noQuadPoints; qp++ ){
    for ( sd=0; sd < noSpaceDim; sd++ ){
      coord(sd) = 0.0;
      for ( sf=0; sf < grid->GetCell(el)->GetSize(); sf++ ){
		  n = grid->GetNode(c->GetNode(sf));
		  coord(sd) += quadpoints(qp,sd) * ( n->GetCoord(sd) );
      }
    }
    integral += Eval(coord,el) * quadweights(qp) * determinant;
  }

  return 0;
}
//-----------------------------------------------------------------------------
real Field::Wkmax_seminorm( int k )
{

  real maxnorm = 0.0;

  MV_ColMat<real> nodalGradient;
  MV_Vector<int> noNeighElms;
  int noNodesInElm;
  Field gradientField(&(*grid));
  MV_Vector<real> loc2Der(noSpaceDim);

  int el,nod,sd,dim;	
  switch( k ){ 
  case 0:
    maxnorm = Lmax_norm();
    break;
  case 1:
    for ( el=0; el < noElms; el++ ){
      CompGradient(el);
      for ( sd=0; sd < noSpaceDim; sd++ ){
	if ( fabs(locgrad(sd)) > maxnorm ) maxnorm = fabs(locgrad(sd));
      }
    }
    break;
  case 2:
    nodalGradient.newsize(noNodes,noSpaceDim);
    nodalGradient = 0.0;
    noNeighElms.newsize(noNodes);
    noNeighElms = 0;
    for ( el=0; el < noElms; el++ ){
      CompGradient(el);
      noNodesInElm = grid->GetCell(el)->GetSize();
      for ( nod=0; nod < noNodesInElm; nod++ ){
	for ( sd=0; sd < noSpaceDim; sd++ ){
	  nodalGradient(grid->GetCell(el)->GetNode(nod),sd) += locgrad(sd);
	}
	noNeighElms(grid->GetCell(el)->GetNode(nod)) += 1;
      }
    }
    for ( nod=0; nod < noNodes; nod++ ){
      for ( sd=0; sd < noSpaceDim; sd++ ){
	nodalGradient(nod,sd) /= noNeighElms(nod);
      }
    }            
    for ( dim=0; dim < noSpaceDim; dim++ ){
      gradientField.InitByNodes(&nodalGradient,dim);
      for ( el=0; el < noElms; el++ ){
	gradientField.EvalGradient(el,loc2Der);
	for ( sd=0; sd < noSpaceDim; sd++ ){
	  if ( fabs(loc2Der(sd)) > maxnorm ) maxnorm = fabs(loc2Der(sd));
	}
      }
    }
    break;
  default:
    cout << "ERROR: Not implemented for this k-order" << endl;
    exit(1);
    break; 
  }
  
  return maxnorm;
}
//-----------------------------------------------------------------------------
real Field::Wkmax_norm( int k )
{

  real maxnorm = 0.0;

  MV_ColMat<real> nodalGradient;
  MV_Vector<int> noNeighElms;
  int noNodesInElm;
  Field gradientField(&(*grid));
  MV_Vector<real> loc2Der(noSpaceDim);

  int el,nod,sd,dim;	
  switch( k ){ 
  case 0:
    maxnorm = Lmax_norm();
    break;
  case 1:
    for ( el=0; el < noElms; el++ ){
      CompGradient(el);
      for ( sd=0; sd < noSpaceDim; sd++ ){
	if ( fabs(locgrad(sd)) > maxnorm ) maxnorm = fabs(locgrad(sd));
      }
    }
    if ( Lmax_norm() > maxnorm ) maxnorm = Lmax_norm();    
    break;
  case 2:
    nodalGradient.newsize(noNodes,noSpaceDim);
    nodalGradient = 0.0;
    noNeighElms.newsize(noNodes);
    noNeighElms = 0;
    for ( el=0; el < noElms; el++ ){
      CompGradient(el);
      noNodesInElm = grid->GetCell(el)->GetSize();
      for ( nod=0; nod < noNodesInElm; nod++ ){
	for ( sd=0; sd < noSpaceDim; sd++ ){
	  nodalGradient(grid->GetCell(el)->GetNode(nod),sd) += locgrad(sd);
	}
	noNeighElms(grid->GetCell(el)->GetNode(nod)) += 1;
      }
    }
    for ( nod=0; nod < noNodes; nod++ ){
      for ( sd=0; sd < noSpaceDim; sd++ ){
	nodalGradient(nod,sd) /= noNeighElms(nod);
      }
    }            
    for ( dim=0; dim < noSpaceDim; dim++ ){
      gradientField.InitByNodes(&nodalGradient,dim);
      for ( el=0; el < noElms; el++ ){
	gradientField.EvalGradient(el,loc2Der);
	for ( sd=0; sd < noSpaceDim; sd++ ){
	  if ( fabs(loc2Der(sd)) > maxnorm ) maxnorm = fabs(loc2Der(sd));
	}
      }
    }
    if ( Wkmax_norm(1) > maxnorm ) maxnorm = Wkmax_norm(1);    
    break;
  default:
    cout << "ERROR: Not implemented for this k-order" << endl;
    exit(1);
    break; 
  }
  
  return maxnorm;
}
//-----------------------------------------------------------------------------
real Field::Lmax_norm()
{

  real maxnorm = 0.0;

  int i;	
  switch( rep ){ 
  case NODAL_VECTOR: 
    for ( i=0; i < noNodes; i++ ){
      if ( fabs((*nodalValuesVector)(i)) > maxnorm ) maxnorm = fabs((*nodalValuesVector)(i));
    }
    break;
  case NODAL_MATRIX: 
    for ( i=0; i < noNodes; i++ ){
      if ( fabs((*nodalValuesMatrix)(i,component)) > maxnorm ) maxnorm = fabs((*nodalValuesMatrix)(i,component));
    }
    break;
  case ELEMENT_VECTOR: 
    for ( i=0; i < noElms; i++ ){
      if ( fabs((*elementValuesVector)(i)) > maxnorm ) maxnorm = fabs((*elementValuesVector)(i));
    }
    break;
  case ELEMENT_MATRIX: 
    for ( i=0; i < noElms; i++ ){
      if ( fabs((*elementValuesMatrix)(i,component)) > maxnorm ) maxnorm = fabs((*elementValuesMatrix)(i,component));
    }
    break;
  case CONSTANT:
    maxnorm = (*constant);
    break;
  case NONE:
    cout << "ERROR: Field is not initialized" << endl;
    exit(1);
    break;
  default:
    cout << "ERROR: Field representation not implemented" << endl;
    exit(1);
    break; 
  }
  
  return maxnorm;
}
//-----------------------------------------------------------------------------
real Field::Wkp_norm( int k, int p )
{

  real norm;
  switch( k ){ 
  case 0:
    norm = Lp_norm(p);
    break;
  case 1:
    norm = pow( pow(Lp_norm(p),real(p)) + pow(Wkp_seminorm(1,p),real(p)), 1.0/real(p) );
    break;
  case 2:
    norm = pow( pow(Wkp_norm(1,p),real(p)) + pow(Wkp_seminorm(2,p),real(p)), 1.0/real(p) );
    break;
  default:
    cout << "ERROR: Not implemented for this k-order" << endl;
    exit(1);
    break; 
  }
  
  return norm;
}
//-----------------------------------------------------------------------------
real Field::Wkp_seminorm( int k, int p )
{

  real norm = 0.0;

  MV_ColMat<real> globalGradient;

  MV_ColMat<real> nodalGradient;
  MV_Vector<int> noNeighElms;
  int noNodesInElm;
  Field gradientField(&(*grid));
  MV_Vector<real> loc2Der(noSpaceDim);

  int el,nod,sd,dim;	
  switch( k ){ 
  case 0:
    norm = Lp_norm(p);
    break;
  case 1:
    globalGradient.newsize(noElms,noSpaceDim);
    for ( el=0; el < noElms; el++ ){
      CompGradient(el);
      for ( sd=0; sd < noSpaceDim; sd++ ){
	globalGradient(el,sd) = locgrad(sd);
      } 
    }     
    for ( sd=0; sd < noSpaceDim; sd++ ){
      gradientField.InitByElms(&globalGradient,sd);
      norm += pow( gradientField.Lp_norm(p), real(p) );
    }
    norm = pow(norm,1.0/real(p));
    break;
  case 2:
    nodalGradient.newsize(noNodes,noSpaceDim);
    nodalGradient = 0.0;
    noNeighElms.newsize(noNodes);
    noNeighElms = 0;
    for ( el=0; el < noElms; el++ ){
      CompGradient(el);
      noNodesInElm = grid->GetCell(el)->GetSize();
      for ( nod=0; nod < noNodesInElm; nod++ ){
	for ( sd=0; sd < noSpaceDim; sd++ ){
	  nodalGradient(grid->GetCell(el)->GetNode(nod),sd) += locgrad(sd);
	}
	noNeighElms(grid->GetCell(el)->GetNode(nod)) += 1;
      }
    }
    for ( nod=0; nod < noNodes; nod++ ){
      for ( sd=0; sd < noSpaceDim; sd++ ){
	nodalGradient(nod,sd) /= noNeighElms(nod);
      }
    }            

    for ( dim=0; dim < noSpaceDim; dim++ ){
      gradientField.InitByNodes(&nodalGradient,dim);
      
      globalGradient.newsize(noElms,noSpaceDim);
      for ( el=0; el < noElms; el++ ){
	gradientField.EvalGradient(el,loc2Der);
	for ( sd=0; sd < noSpaceDim; sd++ ){
	  globalGradient(el,sd) = loc2Der(sd);
	} 
      }     
      for ( sd=0; sd < noSpaceDim; sd++ ){
	gradientField.InitByElms(&globalGradient,sd);
	norm += pow( gradientField.Lp_norm(p), real(p) );
      }
    }
    norm = pow(norm,1.0/real(p));
    break;
  default:
    cout << "ERROR: Not implemented for this k-order" << endl;
    exit(1);
    break; 
  }
  
  return norm;
}
//-----------------------------------------------------------------------------
real Field::Lp_norm( int p )
{

  int el;
  real norm = 0.0;
  for ( el=0; el < noElms; el++ ){
    if ( !UNIFORM_NODAL_BASIS ) GetNodalBasis(el);
    switch( locBasis ){ 
    case LINE_BASIS:
      //LpnormLine(norm,el,real(p));
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    case TRI_BASIS: 
      LpnormTri(norm,el,real(p));
      break;
    case QUAD_BASIS:
      //LpnormQuad(norm,el,real(p));
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    case TET_BASIS:
      LpnormTet(norm,el,real(p));
      break;
    case HEX_BASIS:
      //LpnormHex(norm,el,real(p));
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    case PYR_BASIS:
      //LpnormPyr(norm,el,real(p));
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    case PRISM_BASIS:
      //LpnormPrism(norm,el,real(p));
      cout << "ERROR: Not implemented for this nodal basis" << endl;
      exit(1);
      break;
    default:
      cout << "ERROR: Nodal basis not implemented" << endl;
      exit(1);
      break; 
    }
  }
  norm = pow( norm, 1.0/real(p) );

  return norm;
}
//-----------------------------------------------------------------------------
int Field::LpnormTri( real& norm, int el, real p )
{

  if ( !UNIFORM_NODAL_BASIS ){
    quadrule.init(TRI,quadorder);
    quadrule.getQuadratureRule( quadpoints, quadweights );
    noQuadPoints = quadweights.size();
  }

  determinant = 2.0 * grid->GetCell(el)->ComputeVolume(grid);
  
  Cell *c = grid->GetCell(el);
  Node *n;
  
  int qp,sd,sf;
  for ( qp=0; qp < noQuadPoints; qp++ ){
    for ( sd=0; sd < noSpaceDim; sd++ ){
      coord(sd) = 0.0;
      for ( sf=0; sf < grid->GetCell(el)->GetSize(); sf++ ){
		  n = grid->GetNode(c->GetNode(sf));
		  coord(sd) += quadpoints(qp,sd) * ( n->GetCoord(sd) );
      }
    }
    norm += pow( fabs(Eval(coord,el)), p ) * quadweights(qp) * determinant;
  }

  return 0;
}
//-----------------------------------------------------------------------------
int Field::LpnormTet( real& norm, int el, real p )
{

  if ( !UNIFORM_NODAL_BASIS ){
    quadrule.init(TET,quadorder);
    quadrule.getQuadratureRule( quadpoints, quadweights );
    noQuadPoints = quadweights.size();
  }

  determinant = 6.0 * grid->GetCell(el)->ComputeVolume(grid);

  Cell *c = grid->GetCell(el);
  Node *n;
  
  int qp,sd,sf;
  for ( qp=0; qp < noQuadPoints; qp++ ){
    for ( sd=0; sd < noSpaceDim; sd++ ){
      coord(sd) = 0.0;
      for ( sf=0; sf < grid->GetCell(el)->GetSize(); sf++ ){
		  n = grid->GetNode(c->GetNode(sf));
		  coord(sd) += quadpoints(qp,sd) * ( n->GetCoord(sd) );
      }
    }
    norm += pow( fabs(Eval(coord,el)), p ) * quadweights(qp) * determinant;
  }

  return 0;
}
//-----------------------------------------------------------------------------
int Field::CheckIfInElm( int el )
{

  // This function checks if coord is in el.

  noBasisFcns = grid->GetCell(el)->GetSize();

  if ( !UNIFORM_NODAL_BASIS ) GetNodalBasis(el);
  switch( locBasis ){ 
  case TRI_BASIS: 
    GetBarycentricTri( el );
    break;
  case QUAD_BASIS: 
    if ( GetBarycentricQuad(el) < 0 ){ 
      return 0;
    } else{ 
      return 1;
    }
    break;
  case TET_BASIS: 
    GetBarycentricTet( el );
    break;
  case HEX_BASIS: 
    GetBarycentricHex( el );
    break;
  default:
    cout << "ERROR: The nodal basis is not implemented" << endl;
    exit(1);
  }

  int i;
  for ( i=0; i < noBasisFcns; i++ ){
    if ( barycoord(i) < 0 ) return 0;
  }

  return 1;
}
//-----------------------------------------------------------------------------
int Field::GetBarycentric( int el )
{

  noBasisFcns = grid->GetCell(el)->GetSize();

  if ( !UNIFORM_NODAL_BASIS ) GetNodalBasis(el);
  switch( locBasis ){ 
  case TRI_BASIS: 
    GetBarycentricTri( el );
    break;
  case QUAD_BASIS: 
    GetBarycentricQuad( el );	
    break;
  case TET_BASIS: 
    GetBarycentricTet( el );
    break;
  case HEX_BASIS: 
    GetBarycentricHex( el );
    break;
  default:
    cout << "ERROR: The nodal basis is not implemented" << endl;
    exit(1);
  }

  return 0;
}
//-----------------------------------------------------------------------------
int Field::GetBarycentricTri( int el )
{
  int i;
  for ( i=0; i < grid->GetCell(el)->GetSize(); i++ ) locnodes(i) = grid->GetCell(el)->GetNode(i);

  Node *n1 = grid->GetNode(locnodes(0));
  Node *n2 = grid->GetNode(locnodes(1));
  Node *n3 = grid->GetNode(locnodes(2));
  
  real x1 = n1->GetCoord(0);
  real x2 = n2->GetCoord(0);
  real x3 = n3->GetCoord(0);
  real y1 = n1->GetCoord(1);
  real y2 = n2->GetCoord(1);
  real y3 = n3->GetCoord(1);

  real dN2x = y3-y1;  
  real dN2y = x1-x3;  
  real dN3x = y1-y2;  
  real dN3y = x2-x1;

  real det = dN3y*dN2x - dN3x*dN2y;  
  real d = 1/det;
  
  dN2x *= d;  
  dN2y *= d;  
  dN3x *= d;  
  dN3y *= d;
  real dN1x = - dN2x - dN3x;  
  real dN1y = - dN2y - dN3y;
  
  barycoord.newsize(grid->GetCell(el)->GetSize());
  barycoord(0) = 1 + dN1x*(coord(0)-x1) + dN1y*(coord(1)-y1);
  barycoord(1) = dN2x*(coord(0)-x1) + dN2y*(coord(1)-y1);
  barycoord(2) = dN3x*(coord(0)-x1) + dN3y*(coord(1)-y1);

  return 0;
}
//-----------------------------------------------------------------------------
int Field::GetBarycentricQuad( int el )
{

  MV_ColMat<real> coor(4,2);

  // ATTENTION: In the functions .getBarycentricQuad() & .canBeInsideBox(), 
  //            the numbering of the local nodes differs from the rest of the code:
  //
  // Here:      3----4      Elsewhere:      4----3 
  //            |    |                      |    |
  //            |    |                      |    | 
  //            1----2                      1----2  

  display->InternalError("Field::GetBarycentricQuad()","Quads not implemented.");
  
  // real x1 = n1->GetCoord(0);
//   real x2 = n2->GetCoord(0);
//   real x3 = n3->GetCoord(0);
//   real y1 = n1->GetCoord(1);
//   real y2 = n2->GetCoord(1);
//   real y3 = n3->GetCoord(1);
  
//   coor(0,0) = grid->GetCoor(grid->GetGlobalNode(el,0),0);
//   coor(0,1) = grid->GetCoor(grid->GetGlobalNode(el,0),1);
//   coor(1,0) = grid->GetCoor(grid->GetGlobalNode(el,1),0);
//   coor(1,1) = grid->GetCoor(grid->GetGlobalNode(el,1),1);
//   coor(2,0) = grid->GetCoor(grid->GetGlobalNode(el,3),0);
//   coor(2,1) = grid->GetCoor(grid->GetGlobalNode(el,3),1);
//   coor(3,0) = grid->GetCoor(grid->GetGlobalNode(el,2),0);
//   coor(3,1) = grid->GetCoor(grid->GetGlobalNode(el,2),1);

//   MV_Vector<real> globpt(2);
//   globpt(0) = coord(0);
//   globpt(1) = coord(1);

//   if( CanBeInsideBox(globpt,coor,4,0) == -1 ){ // straight sides
//     return 1;
//   }

//   // solve for local point, using Newton's method
//   const int  max_iterations = 20;    // max no of Newton iterations
//   const real eps = 1.0e-10;          // max error in Newton's method
//   int niter = 0;
//   real error = 1e+20;
  
//   // start point is the centroid of the element, here (0,0)
  
//   real x1 = coor(0,0), y1 = coor(0,1);
//   real x2 = coor(1,0), y2 = coor(1,1);
//   real x3 = coor(2,0), y3 = coor(2,1);
//   real x4 = coor(3,0), y4 = coor(3,1);
  
//   real x21 = x2-x1,  x43 = x4-x3,  x31 = x3-x1,  x42 = x4-x2;
//   real y21 = y2-y1,  y43 = y4-y3,  y31 = y3-y1,  y42 = y4-y2;
  
//   real f1 = 4*globpt(0) - (x1+x2+x3+x4);
//   real f2 = 4*globpt(1) - (y1+y2+y3+y4);
  
//   real j11 = x21 + x43,  j12 = x31 + x42;
//   real j21 = y21 + y43,  j22 = y31 + y42;
  
//   real invdet = 1./(j11*j22 - j12*j21);
  
//   real r,s,r0,s0,dr,ds;
//   r0 = r = (j22*f1 - j12*f2)*invdet;
//   s0 = s = (j11*f2 - j21*f1)*invdet;
  
//   f1 = x43 - x21;
//   f2 = y43 - y21;
  
//   real d1 = (j22*f1 - j12*f2)*invdet;
//   real d2 = (j11*f2 - j21*f1)*invdet;

//   while ( niter++ < max_iterations && error > eps ){
//     dr = r0/( 1 + d1*s ) - r;
//     r += dr;
//     ds = s0/( 1 + d2*r ) - s;
//     s += ds;
//     error = sqr(dr) + sqr(ds);
//     //    cout<<r<<" "<<s<<" "<<error<<endl;
//   }
//   if(niter >= max_iterations){
//     //    warningFP("ElmB4n2d:: findLocPt","Iteration did not converge");
//     cout << "ERROR: Newton iteration didn't converge!!!" << endl; 
//     cout <<" iter="<<niter<<" dr="<<dr<<" ds="<<ds<<endl;
//     cout <<" d1="<<d1<<" d2="<<d1<<" r0="<<r0<<" s0="<<s0<<endl;
//     //coor.print(s_o);
//   }
//   // -- check if locpt is valid 
//   // -- locpt.redim(2);
//   // -- locpt(1) = r;
//   // -- locpt(2) = s;

//   barycoord.newsize(4);
//   barycoord(0) = (0.5-0.5*r)*(0.5-0.5*s);
//   barycoord(1) = (0.5+0.5*r)*(0.5-0.5*s);
//   barycoord(2) = (0.5+0.5*r)*(0.5+0.5*s);
//   barycoord(3) = (0.5-0.5*r)*(0.5+0.5*s);

//   error = sqrt(error);
//   if ( r > 1.001 ||  r < -1.001 ||  s > 1.001 ||  s < -1.001 ) return -1;

//   return 0;
}
//-----------------------------------------------------------------------------
int Field::CanBeInsideBox( MV_Vector<real>& globpt, MV_ColMat<real>& coor, int noNodes, int curved )
{

  //  BooLean inside = dpTRUE;
  int inside = 1;
  real min,max,x,diff,gbpt;
  int i,j,h;
  real comparison_tolerance = 0.01;

  //  h = coor.rows();
  for (i = 0; i < noSpaceDim; i++){
      min = max = coor(0,i);
      for (j = 1; j < noNodes; j++)
        {
          x = coor(j,i);
          if (x < min) min = x;
          else
            if (x > max) max = x;
        }
      if ( curved == 1 )
        {
          diff = (max-min)*0.2;
          max += diff;
          min -= diff;
        }
      gbpt = globpt(i);
      if ( max  + comparison_tolerance < gbpt ||
           gbpt + comparison_tolerance < min )
        {
	  //	  inside = dpFALSE;
	  inside = -1;
          break;
        }

    }
  return inside;
}
//-----------------------------------------------------------------------------
int Field::GetBarycentricTet( int el )
{

  int i;
  for ( i=0; i < grid->GetCell(el)->GetSize(); i++ )
	 locnodes(i) = grid->GetCell(el)->GetNode(i);

  Node *n1 = grid->GetNode(locnodes(0));
  Node *n2 = grid->GetNode(locnodes(1));
  Node *n3 = grid->GetNode(locnodes(2));
  Node *n4 = grid->GetNode(locnodes(2));
  
  real x1 = n1->GetCoord(0);
  real x2 = n2->GetCoord(0);
  real x3 = n3->GetCoord(0);
  real x4 = n4->GetCoord(0);
  real y1 = n1->GetCoord(1);
  real y2 = n2->GetCoord(1);
  real y3 = n3->GetCoord(1);
  real y4 = n4->GetCoord(1);
  real z1 = n1->GetCoord(2);
  real z2 = n2->GetCoord(2);
  real z3 = n3->GetCoord(2);
  real z4 = n4->GetCoord(2);
  
  real j11 = x2-x1;  
  real j12 = y2-y1;  
  real j13 = z2-z1;
  real j21 = x3-x1;  
  real j22 = y3-y1;  
  real j23 = z3-z1;
  real j31 = x4-x1;  
  real j32 = y4-y1;  
  real j33 = z4-z1;

  real dN2x = j22*j33-j23*j32;  
  real dN3x = j13*j32-j12*j33;  
  real dN4x = j12*j23-j13*j22;
  real dN2y = j23*j31-j21*j33;  
  real dN3y = j11*j33-j13*j31;  
  real dN4y = j13*j21-j11*j23;
  real dN2z = j21*j32-j22*j31;  
  real dN3z = j12*j31-j11*j32;  
  real dN4z = j11*j22-j12*j21;

  barycoord.newsize(4);
  barycoord(0) = dN1x*(coord(0)-x2) + dN1y*(coord(1)-y2) + dN1z*(coord(2)-z2);
  barycoord(1) = dN2x*(coord(0)-x1) + dN2y*(coord(1)-y1) + dN2z*(coord(2)-z1);
  barycoord(2) = dN3x*(coord(0)-x1) + dN3y*(coord(1)-y1) + dN3z*(coord(2)-z1);
  barycoord(3) = dN4x*(coord(0)-x1) + dN4y*(coord(1)-y1) + dN4z*(coord(2)-z1);

  return 0;
}
//-----------------------------------------------------------------------------
int Field::GetBarycentricHex( int el )
{

  cout << "The nodal basis is not implemented" << endl;
  exit(1);

  return 0;
}
//-----------------------------------------------------------------------------
int Field::SetTime( real t )
{
  
  time = t;

  return 0;
}
//-----------------------------------------------------------------------------
int Field::SetRotAngle( real a )
{
  
  rotAngle = a;

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalWeierstrassField( real& value )
{

  int sd;

  value = WeierstrassFunction( coord(0) );
  for ( sd = 1; sd < noSpaceDim; sd++ ){
    value *= WeierstrassFunction( coord[sd] );
  }
  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalHolderField( real& value )
{

  int sd;

  value = HolderFunction( coord(0) );
  for ( sd = 1; sd < noSpaceDim; sd++ ){
    value *= HolderFunction( coord[sd] );
  }
  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalHolderFieldRot( real& value )
{

  real x1 = cos(rotAngle)*coord(0) + sin(rotAngle)*coord(1);
  real x2 = cos(rotAngle)*coord(1) - sin(rotAngle)*coord(0);

  value = HolderFunction(x1)*HolderFunction(x2);

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalWeierstrassFieldRot( real& value )
{

  real x1 = cos(rotAngle)*coord(0) + sin(rotAngle)*coord(1);
  real x2 = cos(rotAngle)*coord(1) - sin(rotAngle)*coord(0);

  value = WeierstrassFunction(x1)*WeierstrassFunction(x2);

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalTDHolderField( real& value )
{

  int sd;

  value = HolderFunction( coord(0) );
  for ( sd = 1; sd < noSpaceDim; sd++ ){
    value *= HolderFunction( coord[sd] );
  }
  value *= HolderFunction( time );

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalTDWeierstrassField( real& value )
{

  int sd;

  value = WeierstrassFunction( coord(0) );
  for ( sd = 1; sd < noSpaceDim; sd++ ){
    value *= WeierstrassFunction( coord[sd] );
  }
  value *= WeierstrassFunction( time );

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalTDVelXWeierstrassField( real& value )
{

  int sd;

  value = WeierstrassFunction( coord(0) );
  for ( sd = 1; sd < noSpaceDim; sd++ ){
    value *= WeierstrassFunction( coord[sd] );
  }
  value *= WeierstrassFunction( time );
  
  value *= coord(0)*(1.0-coord(0));
  value += sin(pi*coord(0))*cos(pi*coord(1));

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalTDVelYWeierstrassField( real& value )
{

  int sd;

  value = WeierstrassFunction( coord(0) );
  for ( sd = 1; sd < noSpaceDim; sd++ ){
    value *= WeierstrassFunction( coord[sd] );
  }
  value *= WeierstrassFunction( time );
  
  value *= coord(1)*(1.0-coord(1));
  value += - cos(pi*coord(0))*sin(pi*coord(1));

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalTDVelXHolderField( real& value )
{

  int sd;

  value = HolderFunction( coord(0) );
  for ( sd = 1; sd < noSpaceDim; sd++ ){
    value *= HolderFunction( coord[sd] );
  }
  value *= HolderFunction( time );
  
  value *= coord(0)*(1.0-coord(0));
  value += sin(pi*coord(0))*cos(pi*coord(1));

  return 0;
}
//-----------------------------------------------------------------------------
int Field::EvalTDVelYHolderField( real& value )
{

  int sd;

  value = HolderFunction( coord(0) );
  for ( sd = 1; sd < noSpaceDim; sd++ ){
    value *= HolderFunction( coord[sd] );
  }
  value *= HolderFunction( time );
  
  value *= coord(1)*(1.0-coord(1));
  value += - cos(pi*coord(0))*sin(pi*coord(1));

  return 0;
}
//-----------------------------------------------------------------------------
int Field::Eval2DHolderField( real& value )
{

  value += HolderFunction2D( coord(0), coord(1) );

  return 0;
}
//-----------------------------------------------------------------------------
int Field::SetHolderOrder( int order )
{

  N_Holder = order;

  return 0;
}
//-----------------------------------------------------------------------------
real Field::HolderFunction_old( real x )
{

  // Holder functions
  int j,k;
  real f = 0.0;
  real fbase = parameters(0);
  real gamma = parameters(1);
  real delta = parameters(2);

  for ( j=0; j < N_Holder; j++ ){
    for ( k=0; k < pow(2,N_Holder)-1; k++ ){
      f += pow(2.0,-j*(0.5+delta)) * HaarWavelet( j, k, x ); 
    }		     
  }
  return fbase + gamma*f;
}
//-----------------------------------------------------------------------------
real Field::WeierstrassFunction( real x )
{

  int j,k;
  real f = 0.0;
  real fbase = parameters(0);
  real gamma = parameters(1);
  real delta = parameters(2);

  real b = 1.5;

  for ( j=-N_Holder; j < N_Holder; j++ ){
    //for ( j=0; j < N_Holder; j++ ){
    k = int( floor(pow(b,j)) );
    f += pow(b,-j*delta) * ( 1.0 - cos( pow(b,j)*x - k ) );
  }
  return fbase + gamma*f;
}
//-----------------------------------------------------------------------------
real Field::HolderFunction( real x )
{

  // Holder functions
  int j,k;
  real f = 0.0;
  real fbase = parameters(0);
  real gamma = parameters(1);
  real delta = parameters(2);
  real crd;
  for ( j=0; j < N_Holder; j++ ){
    crd = 0.5 - (pow(2.0,j)*x-floor(pow(2.0,j)*x));
    if ( fabs(crd) > 0.0 ){
      f += pow(2.0,-j*delta) * crd/fabs(crd); 
    }
  }		     
  return fbase + gamma*f;
}
//-----------------------------------------------------------------------------
real Field::HolderFunction2D( real x, real y )
{

  int j,kx,ky;

  real fH = 0.0;
  real fV = 0.0;
  real fD = 0.0;
  real fbase = parameters(0);
  real gammaH = fbase * parameters(1);
  real deltaH = parameters(2);
  real gammaV = fbase * parameters(3);
  real deltaV = parameters(4);
  real gammaD = fbase * parameters(5);
  real deltaD = parameters(6);

  for ( j=0; j < N_Holder; j++ ){
    for ( kx=0; kx < pow(2,N_Holder)-1; kx++ ){
      for ( ky=0; ky < pow(2,N_Holder)-1; ky++ ){
	fH += pow(2.0,-j*(1.0+2*deltaH)) * HaarWavelet2DH( j, kx, ky, x, y ); 
	fV += pow(2.0,-j*(1.0+2*deltaV)) * HaarWavelet2DV( j, kx, ky, x, y ); 
	fD += pow(2.0,-j*(1.0+2*deltaD)) * HaarWavelet2DD( j, kx, ky, x, y ); 
      }		     
    }
  }
  return fbase + gammaH*gammaH*fH + gammaV*gammaV*fV + gammaD*gammaD*fD;
}
//-----------------------------------------------------------------------------
real Field::HaarWavelet( int i, int k, real x )
{
  return pow(2.0,0.5*i) * HaarMotherWavelet( pow(2.0,i)*x - k );
}
//-----------------------------------------------------------------------------
real Field::HaarWavelet2DH( int i, int kx, int ky, real x, real y )
{
  return pow(2.0,i) * HaarScaleFcn(pow(2.0,i)*x - kx) * HaarMotherWavelet(pow(2.0,i)*y - ky);
}
//-----------------------------------------------------------------------------
real Field::HaarWavelet2DV( int i, int kx, int ky, real x, real y )
{
  return pow(2.0,i) * HaarMotherWavelet(pow(2.0,i)*x - kx) * HaarScaleFcn(pow(2.0,i)*y - ky) ;
}
//-----------------------------------------------------------------------------
real Field::HaarWavelet2DD( int i, int kx, int ky, real x, real y )
{
  return pow(2.0,i) * HaarMotherWavelet(pow(2.0,i)*x - kx) * HaarMotherWavelet(pow(2.0,i)*y - ky);
}
//-----------------------------------------------------------------------------
real Field::HaarMotherWavelet( real x )
{
  if ( x > 0.0 && x < 0.5 )
    return 1.0;
  else{
    if ( x > 0.5 && x < 1.0 )
      return -1.0;
    else{
      return 0.0;
    }
  }
}
//-----------------------------------------------------------------------------
real Field::HaarScaleFcn( real x )
{
  if ( x > 0.0 && x < 1.0 )
    return 1.0;
  else{
    return 0.0;
  }
}
//-----------------------------------------------------------------------------
real Field::HaarMotherWavelet2DH( real x, real y )
{
  if ( x > 0.0 && x < 1.0 ){
    if ( y > 0.0 && y < 0.5 )
      return 1.0;
    else{
      if ( y > 0.5 && y < 1.0 ){
	return -1.0;
      }
      else{
	return 0.0;
      }
    }
  }
  else{
    return 0.0;
  }
}
//-----------------------------------------------------------------------------
real Field::HaarMotherWavelet2DV( real x, real y )
{

  if ( y > 0.0 && y < 1.0 ){
    if ( x > 0.0 && x < 0.5 )
      return 1.0;
    else{
      if ( x > 0.5 && x < 1.0 ){
	return -1.0;
      }
      else{
	return 0.0;
      }
    }
  }
  else{
    return 0.0;
  }
}
//-----------------------------------------------------------------------------
real Field::HaarMotherWavelet2DD( real x, real y )
{

  if ( y > 0.0 && y < 0.5 && x > 0.0 && x < 0.5 ){
    return 1.0;
  }
  else{
    if ( y > 0.0 && y < 0.5 && x > 0.5 && x < 1.0 ){
      return -1.0;
    }
    else{
      if ( y > 0.5 && y < 1.0 && x > 0.0 && x < 0.5 ){
	return -1.0;
      }
      else{
	if ( y > 0.5 && y < 1.0 && x > 0.5 && x < 1.0 ){
	  return 1.0;
	}
	else{
	  return 0.0;
	}
      }
    }
  }
}
//-----------------------------------------------------------------------------
real Field::HaarScaleFcn2D( real x, real y )
{
  
  if ( x > 0.0 && x < 1.0 && y > 0.0 && y < 1.0 ){
    return 1.0;
  }
  else{
    return 0.0;
  }
}
//-----------------------------------------------------------------------------
int Field::EvalPolynomialField( real& value )
{

  int degree = ( noParameters - 1 ) / noSpaceDim; // Should be an integer
  real val = 0;
  real tmp = 0;
  for ( int i = 1; i < degree ; i++ ){
    for ( int j = 0; j < noSpaceDim; j++ ){
      tmp = parameters( (i-1) * noSpaceDim + j + 1 ) * pow(coord[j],i); 
    }
  }
  value = parameters(0) + tmp;

  return 0;
}
//-----------------------------------------------------------------------------
