// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __FIELD_HH
#define __FIELD_HH

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <mvvtp.h>
#include <mvmtp.h>
#include "Grid.hh"
#include "QuadratureRules.h"

enum FunctionClass { UNDEF, POLYNOMIAL, HOLDERFCN, HOLDERFCNTD, HOLDERFCNTDVELX, HOLDERFCNTDVELY, 
		     HOLDERFCN2D, HOLDERFCNROT, WEIERSTRASS, WEIERSTRASSROT, WEIERSTRASSTDVELX, 
		     WEIERSTRASSTDVELY, WEIERSTRASSFCNTD };

enum representation { NONE, NODAL_VECTOR, NODAL_MATRIX, ELEMENT_VECTOR, ELEMENT_MATRIX, CONSTANT, FUNCTION };
enum nodalbasis { NO_BASIS, LINE_BASIS, TRI_BASIS, QUAD_BASIS, TET_BASIS, HEX_BASIS, PYR_BASIS, 
		  PRISM_BASIS, SIMPLEX_BASIS, HYPERCUBE_BASIS };
 
class Field{
  
public:

  Field();
  Field( Grid* g );

  ~Field(){}

  void operator = ( Field* fld );

  int AttachGrid( Grid* g );
  Grid* GetGrid();
  
  int GetNoSpaceDim(){ return noSpaceDim; }
  int GetNoElms(){ return noElms; }
  int GetNoNodes(){ return noNodes; }

  int InitByNodes( Vector *nodVal, int nsys, int comp );// Initialize field by the nodal values.
  int InitByNodes( MV_Vector<real>* nodVal );           // Initialize field by the nodal values.
  int InitByNodes( MV_ColMat<real>* nodVal, int col );  // ...by the nodal values in col.
  int InitByElms( MV_Vector<real>* elVal );             // ...by the values over the elements. 
  int InitByElms( MV_ColMat<real>* elVal, int col );    // ...by the values over the elements in col. 
  int InitByConstant( real* c );                         // ...by a constant value const. 
  int InitByFcn( MV_Vector<real> par, FunctionClass fcntype ); // ...by a function expression.
  
  int GetNodalBasis( int el );

  void GetLocalNodes(int el, real *local_nodes);

  representation GetRep() { return rep; }              // Returns the representation for the field.
  FunctionClass GetFcn() { return whatFcn; }           // Returns the functionclass for the field.

  int GetElementVector( MV_Vector<real>& elVec );  // These functions returns a MV_Vector containing
  int GetNodalVector( MV_Vector<real>& elVec );    // the element and nodal values repectively.

  real GetElemValue(int el);   // Returns midpoint value for element el
  real GetNodalValue(int nod); // Returns value in node nod.

  real Eval();                                   // Evaluate constant field.
  real Eval( MV_Vector<real>& point );           // Evaluate the field in point.
  real Eval( MV_Vector<real>& point, int el );   // Evaluate the field in point, given 
  real Eval( int el );                           // what element point is in.

  real EvalGradient( int el, int comp );
  int EvalGradient( int el, MV_Vector<real>& grad );

  int Integrate( real& integral );

  real Lp_norm( int p );
  real Lmax_norm();

  real Wkp_seminorm( int k, int p );
  real Wkp_norm( int k, int p );
  real Wkmax_seminorm( int k );
  real Wkmax_norm( int k );
  
  // Utilities for fractal Holder functions
  int SetHolderOrder( int order );
  int SetTime( real t );
  real GetTime(){ return time; }
  int SetRotAngle( real a );
  real GetRotAngle(){ return rotAngle; }


 private:

  Grid* grid;
  int grid_init;

  int noNodes;
  int noElms;
  int noSpaceDim;

  real dN1x,dN1y,dN1z,dN2x,dN2y,dN2z,dN3x,dN3y,dN3z,dN4x,dN4y,dN4z;

  real determinant;
  
  real pi;

  representation rep;
  nodalbasis locBasis;
  FunctionClass whatFcn;                   

  bool UNIFORM_NODAL_BASIS;
  
  int component;
  MV_Vector<real>* nodalValuesVector;      // Nodal vector values
  MV_ColMat<real>* nodalValuesMatrix;      // Nodal matrix values
  MV_Vector<real>* elementValuesVector;    // Element vector values   
  MV_ColMat<real>* elementValuesMatrix;    // Element matrix values   
  real* constant;                          // Constant value
  
  MV_Vector<real> parameters;              // Parameters for functions
  int noParameters;                

  MV_Vector<real> locgrad;
  MV_Vector<real> locnodes;
  
  QuadratureRules quadrule;
  int quadorder;
  int noQuadPoints;
  MV_ColMat<real> quadpoints;
  MV_Vector<real> quadweights;

  MV_Vector<real> coord;
  MV_Vector<real> barycoord;
  int noBasisFcns;

  real Nodes2elms( int el );
  real Elms2nodes( int node );

  int EvalNodalField( real& value );
  int EvalElementField( real& value );
  int EvalConstantField( real& value );
  int EvalNodalField( real& value, int el );
  int EvalElementField( real& value, int el );
  int EvalConstantField( real& value, int el );
  int EvalFunctionField( real& value );

  int CompGradient( int el );
  int CompGradientQuad( int el );     // 2D uniform quadratic grid.
  int CompGradientTri( int el );      // 2D triangular elements (NOT COMPLETE!!
  int CompGradientTet( int el );      // 3D tetrahedron elements.
 
  int IntegrateQuadratureTri( real& integral, int el );
  int IntegrateQuadratureTet( real& integral, int el );

  int LpnormTri( real& pnorm, int el, real p );
  int LpnormTet( real& pnorm, int el, real p );

  int CheckIfInElm( int el );

  int GetBarycentric( int el );
  int GetBarycentricTri( int el );
  int GetBarycentricQuad( int el );
  int CanBeInsideBox( MV_Vector<real>& globpt, MV_ColMat<real>& coor, int noNodes, int curved );
  int GetBarycentricTet( int el );
  int GetBarycentricHex( int el );

  // Holder test functions. ("Maximal wild" Holder functions)

  int N_Holder;
  real time;
  real rotAngle;

  int EvalWeierstrassField( real& value );
  int EvalWeierstrassFieldRot( real& value );
  int EvalTDVelXWeierstrassField( real& value );
  int EvalTDVelYWeierstrassField( real& value );
  int EvalTDWeierstrassField( real& value );

  int EvalHolderField( real& value );
  int EvalHolderFieldRot( real& value );
  int EvalTDHolderField( real& value );
  int EvalTDVelXHolderField( real& value );
  int EvalTDVelYHolderField( real& value );
  int Eval2DHolderField( real& value );

  int EvalPolynomialField( real& value );

  real WeierstrassFunction( real x );

  real HolderFunction_old( real x );
  real HolderFunction( real x );
  real HolderFunction2D( real x, real y );
  
  real HaarWavelet( int i, int k, real x );
  real HaarMotherWavelet( real x );
  real HaarScaleFcn( real x );

  real HaarWavelet2DH( int i, int kx, int ky, real x, real y );
  real HaarWavelet2DV( int i, int kx, int ky, real x, real y );
  real HaarWavelet2DD( int i, int kx, int ky, real x, real y );
  real HaarScaleFcn2D( real x, real y );

  real HaarMotherWavelet2DH( real x, real y );
  real HaarMotherWavelet2DV( real x, real y );
  real HaarMotherWavelet2DD( real x, real y );
  
};

#endif





