// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "GlobalField.hh"
#include <Display.hh>
#include <Settings.hh>
#include <Grid.hh>
#include <Vector.hh>

extern Settings *settings;

//-----------------------------------------------------------------------------
GlobalField::GlobalField()
{
  nvd = 0;
  nodal_values = 0;
  cell_values = 0;

  rep = NONE;
  grid_init = false;
};
//-----------------------------------------------------------------------------
GlobalField::GlobalField( Grid* g )
{
  nvd = 0;
  nodal_values = 0;
  cell_values = 0;

  rep = NONE;

  grid = g;
  grid_init = true;

  no_nodes = grid->GetNoNodes();
  no_cells = grid->GetNoCells();

  settings->Get("space dimension",&nsd);

  uniform_nodal_basis = false;
  if ( (*g).AllCellsEqual() ) uniform_nodal_basis = true;
}
//-----------------------------------------------------------------------------
GlobalField::~GlobalField()
{
};
//-----------------------------------------------------------------------------
int GlobalField::GetNoCells()
{
  return no_cells;
}
//-----------------------------------------------------------------------------
int GlobalField::GetNoNodes()
{
  return no_nodes;
}
//-----------------------------------------------------------------------------
int GlobalField::GetVectorDim(int cell)
{
  return nvd;
}
//-----------------------------------------------------------------------------
void GlobalField::AttachGrid(Grid* g)
{
  grid = g;
  grid_init = true;

  no_nodes = grid->GetNoNodes();
  no_cells = grid->GetNoCells();

  settings->Get("space dimension",&nsd);
  
  uniform_nodal_basis = false;
  if ( (*g).AllCellsEqual() ) uniform_nodal_basis = true;
}
//-----------------------------------------------------------------------------
int GlobalField::GetNodalBasis(int cell)
{
  no_shape_functions = grid->GetCell(cell)->GetSize();
  
  if ( nsd == 1 ){
    cout << "ERROR: Not implemented for 1d" << endl;
    exit(1);
  }
  if ( nsd == 2 ){
    switch(no_shape_functions){ 
    case 3: 
      locBasis = TRI_BASIS;
      break;
    case 4:
      locBasis = QUAD_BASIS;
      break;
    default:
      display->InternalError("GlobalField::GetNodalBasis","The nodal basis is not implemented");
    }
  }  
  if ( nsd == 3 ){
    switch(no_shape_functions){ 
    case 4: 
      locBasis = TET_BASIS;
      break;
    case 8:
      locBasis = HEX_BASIS;
      break;
    default:
      display->InternalError("GlobalField::GetNodalBasis","The nodal basis is not implemented");
    }
  }  

  return 0;
}
//-----------------------------------------------------------------------------
void GlobalField::GetLocalNodes(int cell, real *local_nodes)
{
  GetLocalNodes(cell,0,local_nodes);
}
//-----------------------------------------------------------------------------
void GlobalField::GetLocalNodes(int cell, int component, real *local_nodes)
{

  switch( rep ){ 
  case NODAL: 
    for (int i=0; i < no_shape_functions; i++){ 
      local_nodes[i] = nodal_values->Get(grid->GetCell(cell)->GetNode(i)*nvd + component); 
    }
    break;
  case CELL:
    for (int i=0; i < no_shape_functions; i++) 
      local_nodes[i] = cell_values->Get(cell*nvd + component); 
    break;
  case CONSTANT:
    for (int i=0; i < no_shape_functions; i++) 
      local_nodes[i] = constant;
    break;
  case NONE:
    display->InternalError("GlobalField::GetLocalNodes()","GlobalField is not initialized");
  default:
    display->InternalError("GlobalField::GetLocalNodes()","Wrong GlobalField representation");
  }  
}
//-----------------------------------------------------------------------------
void GlobalField::InitByNodes(Vector *nodal_values, int nvd)
{
  this->nodal_values = nodal_values;
  this->nvd = nvd;
  rep = NODAL;
}
//-----------------------------------------------------------------------------
void GlobalField::InitByCells(Vector *cell_values, int nvd)
{
  this->cell_values = cell_values;
  this->nvd = nvd;
  rep = CELL;
}
//-----------------------------------------------------------------------------    
void GlobalField::InitByConstant(real c)
{
  constant = c;
  rep = CONSTANT;
}
//-----------------------------------------------------------------------------
real GlobalField::GetConstantValue()
{
  switch( rep ){ 
  case CONSTANT:
    return constant;
    break;
  default:
    display->InternalError("GlobalField::GetLocalNodes()","Not a constant GlobalField");
  }
}
//-----------------------------------------------------------------------------
real GlobalField::GetNodalValue(int node)
{
  return GetNodalValue(node,0);
}
//-----------------------------------------------------------------------------
real GlobalField::GetNodalValue(int node, int component)
{
  switch( rep ){ 
  case NODAL: 
    return nodal_values->Get(node*nvd+component);
    break;
  case CONSTANT:
    return constant;
    break;
  default:
    display->InternalError("GlobalField::GetLocalNodes()","GlobalField is not initialized");
  }
}
//-----------------------------------------------------------------------------
real GlobalField::GetCellValue(int cell)
{
  return GetCellValue(cell,0);
}
//-----------------------------------------------------------------------------
real GlobalField::GetCellValue(int cell, int component)
{
  switch( rep ){ 
  case CELL:
    return cell_values->Get(cell*nvd+component);
    break;
  case CONSTANT:
    return constant;
    break;
  default:
    display->InternalError("GlobalField::GetLocalNodes()","GlobalField is not initialized");
  }
}
//-----------------------------------------------------------------------------
