// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __GRID_RELATION_HH
#define __GRID_RELATION_HH

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "kw_constants.h"
#include "Field.hh"
#include "SparseMatrix.hh"
#include "Settings.hh"

extern Settings *settings;

class GridRelation{
  
 public: 

  GridRelation();
  GridRelation( Grid *g1, Grid *g2 );
  
  void operator = (GridRelation& gr);
  
  int proj2PWConst( Field& f );
  int trans2PWConst( MV_Vector<real>& fineVal, MV_Vector<real>& coarseVal );
  int proj2PWLin( Field& f );
  // These functions project the field definded on the fine grid onto the 
  // space of piecewise constant/linear functions on the coarse grid 
  // respectively.

  int proj2PWConst( MV_Vector<real>& fineVal, MV_Vector<real>& coarseVal );
  // This function projects the element values on the fine grid (fineVal) 
  // onto the space of piecewise constant function on the coarse grid (coarseVal). 

  int cpFieldRep( Field& field, Field& cpfield );
  // This function copies a Field representation on one grid onto a nodal 
  // Field representation on the other grid.

  int initNestElMap();
  // This function creates prolongation and restriction matrices, for 
  // the nested element mapping.
  int initNestElProlongation();
  int initNestElRestriction();

  int initNodalMap();
  // This function creates prolongation and restriction matrices, for 
  // the nodal mapping.
  int initNodalProlongation();
  int initNodalRestriction();

  int projNodalData( MV_Vector<real> &val, MV_Vector<real> &projval );  
  int projNodalData( MV_ColMat<real> &val, MV_ColMat<real> &projval );  
  int projNodalData( MV_Vector<real> &val );  
  int projNodalData( MV_ColMat<real> &val );  
  // This function projects the nodal values from val to projval.
  // The function can handle grids where one grid is finer than the other, 
  // or where the grids are approximately of the same size. It cannot 
  // handle (1) a situation where grid1 is finer than grid2 in one part of 
  // the grid, and grid2 is finer in another part. Also it cannot handle 
  // (2) a situation where the grids have not the same boundary.

  int projNestedElData( MV_ColMat<real> &val, MV_ColMat<real> &projval );  
  int projNestedElData( MV_Vector<real> &val, MV_Vector<real> &projval ); 
  int projNestedElData( MV_ColMat<real> &val );  
  int projNestedElData( MV_Vector<real> &val );
  // This function projects element data from elval to projelval, where 
  // the two grids that are nested.

  int avNodesOverElms( MV_Vector<real> &nodVal, MV_Vector<real> &elVal );  
  int avNodesOverElms( MV_Vector<real> &Val );  
  // These functions computes the average of the nodal vectors over the elements.
  // In the first function you supply the element vector elVal, where the 
  // average of nodVal is stored. In the second function the vector Val is 
  // returned containing the element values for the average of the supplied 
  // nodal vector Val.

  int writeProlongationMatrix( char *file );

  int GetNoNodes1();
  int GetNoNodes2();
  int GetNoCells1(); 
  int GetNoCells2(); 
  int GetNoSpaceDim();
  int GetFineNoNodes();
  int GetFineNoElms();
  int GetCoarseNoNodes();
  int GetCoarseNoElms();

  bool NodalProlongationInitialized();
  bool NodalRestrictionInitialized();
  bool ElementProlongationInitialized();
  bool ElementRestrictionInitialized();

  Grid* GetGrid1() { return grid1; }
  Grid* GetGrid2() { return grid2; }

  Grid* GetCoarseGrid();
  Grid* GetFineGrid();
  
  ~GridRelation();
  
 private:
  
  Grid *grid1;
  Grid *grid2;  

  bool prol_init;
  bool res_init;
  bool el_prol_init;
  bool el_res_init;

  int noNodes1;
  int noNodes2;
  int noElms1; 
  int noElms2; 
  int nsd;

  int fineNoNodes;
  int coarseNoNodes;
  int fineNoElms;
  int coarseNoElms;
  int finegrid;
  int coarsegrid;
  int avNoNodesInElm;

  int allElmsOfSameType;

  MV_Vector<real> baryCoord;

  SparseMatrix *prolMat;
  SparseMatrix *resMat;

  SparseMatrix *elProlMat;
  SparseMatrix *elResMat;

  int restriction();
  int prolongation();
  // The function .initNodalMap() calls .prolongation() to create the prolongation 
  // matrix prolMat. Then .initNodalMap calls restriction to construct the 
  // restriction matrix resMat from prolMat.

  int elementProlongation();
  int elementRestriction();
  // The function .initNestedELMap() calls .elementProlongation() to create the 
  // prolongation matrix elProlMat. Then .initNestedELMap calls .elementRestriction 
  // to construct the restriction matrix elResMat from elProlMat.

  int whatElement( MV_Vector<real>& point, int& el, int whatGrid );
  // The function whatElement() returns the element point is in. It also sets baryCoord 
  // equal to the barycentric coordinates for point relative the element point is in.
  // The input el, specifies the previous taken element. 
  // The function starts to check el and the neighbors of el.

  int nextNode( int& node, MV_Vector<int>& nodesTaken, MV_Vector<int>& nodeData, int whatGrid );
  // nextNode selects the next node in whatGrid as nodeData(0), and places 
  // its neighbors in nodeData(1)-nodeData(noNeighbors), further it set  
  // nodeData(nodeData.size()-1) = noNeighbors. The function start searching the neighbors
  // of prevNode, and check so that the node is not already taken.

  int nextEl( int& el, MV_Vector<int>& elmsTaken, MV_Vector<int>& elNeighData, int whatGrid );
  // nextEl selects the next element in whatGrid as elData(0), and places 
  // its neighbors in elData(1)-elData(noNeighbors), further it set  
  // elData(elData.size()-1) = noNeighbors. The function start searching the neighbors
  // of prevEl, and check so that the element is not already taken.

  int nodeNeighbors1( MV_Vector<int>& neighData );
  int nodeNeighbors2( MV_Vector<int>& neighData );
  // These functions find the neighbors to the input node neighData(0), 
  // in grid1 or grid2 respectively. Then returns the neighbors in 
  // neighData(0)-neighData(noNeighbors), where neighData(neighData.size()-1) = noNeighbors. 

  int elementNeighbors1( MV_Vector<int>& neighborElms );
  int elementNeighbors2( MV_Vector<int>& neighborElms );
  // This function finds the neighbor elements to neighborElms(0),
  // and returns the neighbors in neighborElms(0)-neighborElms(noNeighbors),
  // where neighborElms(neighborElms.size()-1) = noNeighbors.

  int checkIfInElm( MV_Vector<real>& point, int el, int whatGrid );
  // This function returns 1 if node is in el, and 0 else. 

  int GetBarycentric( MV_Vector<real>& point, int el, int whatGrid );
  // This function chooses one of the functions below, and returns the number 
  // of barycentric coordinates for point relative the element el. 
  // Also it saves these barycentric coordinates in the vector baryCoord.

  int GetBarycentricTri( MV_Vector<real>& point, int el, int whatGrid );
  int GetBarycentricTet( MV_Vector<real>& point, int el, int whatGrid );
  int GetBarycentricQuad( MV_Vector<real>& point, int el, int whatGrid );
  int canBeInsideBox( MV_Vector<real>& globpt, MV_ColMat<real>& coor, int noNodes, int curved );
  int GetBarycentricHex( MV_Vector<real>& point, int el, int whatGrid );
 

  int updateProl( int node, int el );
  // This function updates the prolongation matrix. 
  // OBSERVE: This function must be modified for hybrid grids.

};

#endif

/*
  LIMITATIONS:
  
  To use hybrid grids one has to modify the allocation of the sparse matrices prolMat 
  and resMatand the vectors prolCoeff and resCoeff. 

  The class can handle grids where one grid is finer than the other, or where the 
  grids are approximately of the same size. It cannot handle (1) a situation where 
  grid1 is finer than grid2 in one part of the grid, and grid2 is finer in another 
  part. Also it cannot handle (2) a situation where the grids does not have the same 
  boundary.

  Some sizes of vectors are set static to such values that are sufficient for the elements 
  now used in the code, e.g. sizes of vectors containing element neighbors are set to 100, 
  which is enough for 3D tetrahedrons. But this might have to be changed for more general 
  elements and higher dimensions.
  

--------------------------------------------------------------------------------------------

  Example 1: (Project the nodal vector nodVal on grid1 to the nodal vector projVal on grid2)

  Grid grid1;
  Grid grid2;

  GridRelation gridrel(grid1,grid2);

  gridrel.initNodalMap();

  MV_Vector<real> projVal;

  gridrel.projNodalData(nodVal,projVal);
  
--------------------------------------------------------------------------------------------

  Example 2: (Project element values on a finer grid fineElVal onto the space of 
             piecewise constant functions on a coarser grid.)
  
  Grid finegrid;
  Grid coarsegrid;

  GridRelation gridrel(finegrid,coarsegrid);
  
  gridrel.initNestedElData;

  MV_Vector<real> coarseElVal;

  gridrel.proj2PWConst(fineElVal,coarseElVal);

--------------------------------------------------------------------------------------------


  Written by Johan Hoffman, Chalmers Finite Element Center 1999 (c).
*/







