// Original author: Johan Hoffman. Added to Kraftwerk 2002-02-05.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "LocalField.hh"
#include "GlobalField.hh"
#include "FunctionSpace.hh"
#include "FiniteElement.hh"
#include "ShapeFunction.hh"
#include <Settings.hh>

extern Settings *settings;

//-----------------------------------------------------------------------------
LocalField::LocalField()
{
  dim = 0;
}
//-----------------------------------------------------------------------------
LocalField::LocalField(FiniteElement *element)
{
  functionspace = element->GetFunctionSpace();

  dim = functionspace->GetDim();

  dof_values = new real[dim];
  gradient   = new real[functionspace->GetSpaceDim()];

  gradient_computed = false;
}
//-----------------------------------------------------------------------------
LocalField::~LocalField()
{
  delete dof_values;
  delete gradient;
}
//----------------------------------------------------------------------------- 
void LocalField::AttachFiniteElement(FiniteElement *element)
{
  functionspace = element->GetFunctionSpace();

  dim = functionspace->GetDim();
  
  dof_values = new real[dim];
  gradient   = new real[functionspace->GetSpaceDim()];
  
  gradient_computed = false;
}
//----------------------------------------------------------------------------- 
void LocalField::operator = ( LocalField &lf )
{
  if ( functionspace != lf.GetFunctionSpace() ){
    functionspace = lf.GetFunctionSpace();
    dim = functionspace->GetDim();
    dof_values = new real[dim];
    gradient   = new real[functionspace->GetSpaceDim()];
  }
  
  gradient_computed = false;
} 
//-----------------------------------------------------------------------------
void LocalField::Add(LocalField &lf)
{
  if ( functionspace != lf.GetFunctionSpace() )
    display->InternalError("LocalField::Add()","Cannot add LocalFields from different functionspaces");

  for (int i=0;i<dim;i++) dof_values[i] += lf.GetDofValue(i);
}
//-----------------------------------------------------------------------------
void LocalField::Mult(real a)
{
  for (int i=0;i<dim;i++) dof_values[i] *= a;
}
//-----------------------------------------------------------------------------
const ShapeFunction* LocalField::GetShapeFunction(int i) const 
{
  return functionspace->GetShapeFunction(i);
}
//-----------------------------------------------------------------------------
real LocalField::GetCoord(int node, int dim) const
{
  return functionspace->GetCoord(node,dim);
}
//-----------------------------------------------------------------------------
real LocalField::GetDofValue(int dof) const
{
  return dof_values[dof];
}
//-----------------------------------------------------------------------------
real LocalField::GetMeanValue() const
{
  real mean = 0.0;
  for (int i=0;i<dim;i++) mean += dof_values[i];

  return mean/dim;
}
//-----------------------------------------------------------------------------
int LocalField::GetDim() const
{
  functionspace->GetDim();
}
//-----------------------------------------------------------------------------
int LocalField::GetCellNo() const
{
  return functionspace->GetCellNo();
}
//-----------------------------------------------------------------------------
void LocalField::Refill(int c_no, int component, GlobalField *field)
{
  field->GetLocalNodes(c_no,component,dof_values);

  functionspace->Refill(c_no);

  gradient_computed = false;
}
//-----------------------------------------------------------------------------
void LocalField::Refill(int c_no, real *nvals)
{
  // FIXME: There is no check for the size of nvals.

  for (int i=0; i < functionspace->GetDim(); i++) dof_values[i] = nvals[i];
  
  functionspace->Refill(c_no);

  gradient_computed = false;
}
//-----------------------------------------------------------------------------
FunctionSpace* LocalField::GetFunctionSpace()
{
  return functionspace;
}
//-----------------------------------------------------------------------------
FiniteElement* LocalField::GetFiniteElement() const
{
  return functionspace->GetFiniteElement();
}
//-----------------------------------------------------------------------------
void LocalField::ComputeGradient()
{
  EvalGradient(gradient);
}
//-----------------------------------------------------------------------------
void LocalField::EvalGradient(real *grad)
{
  for (int j=0;j<functionspace->GetSpaceDim();j++){
    grad[j] = 0.0;
    for (int i=0;i<functionspace->GetDim();i++) 
      grad[j] += dof_values[i]*functionspace->GetGradient(i,j);
  }
  gradient_computed = true;
}
//-----------------------------------------------------------------------------
real LocalField::EvalGradient(int component)
{
  if ( !gradient_computed ) ComputeGradient();
  return gradient[component];
}
//-----------------------------------------------------------------------------
real LocalField::dx()
{
  if ( !gradient_computed ) ComputeGradient();
  return gradient[0];
}
//-----------------------------------------------------------------------------
real LocalField::dy()
{
  if ( !gradient_computed ) ComputeGradient();
  return gradient[1];
}
//-----------------------------------------------------------------------------
real LocalField::dz()
{
  if ( !gradient_computed ) ComputeGradient();
  return gradient[2];
}
//-----------------------------------------------------------------------------
