#include "ScaleExtrapolation.hh"

MV_Vector<real> ScaleExtrapolation::ComputeCovariance( MV_Vector<real>* v, MV_Vector<real>* w, GridRelation* gridRel ){

  // this function computes the covariance [vw]^h-[v]^h[w]^h

  if ( v->size() != w->size() ){
    cout << "ERROR: Wrong size of vectors" << endl;
    exit(1);
  }

  MV_Vector<real> vw;
  MV_Vector<real> v_h;
  MV_Vector<real> w_h;
  MV_Vector<real> vw_h;
  MV_Vector<real> cov;

  int fineNoNodes = gridRel->GetFineNoNodes();
  int coarseNoNodes = gridRel->GetCoarseNoNodes();
  int fineNoElms = gridRel->GetFineNoElms();
  int coarseNoElms = gridRel->GetCoarseNoElms();

  int i;
 
  if ( v->size() == gridRel->GetFineNoNodes() ){

    if ( !gridRel->NodalRestrictionInitialized() ) gridRel->initNodalRestriction();

    vw.newsize(fineNoNodes);
    cov.newsize(coarseNoNodes);

    for ( i=0; i < fineNoNodes; i++ ) vw(i) = (*v)(i) * (*w)(i);
    
    
    gridRel->projNodalData(*v,v_h);
    gridRel->projNodalData(*w,w_h);
    gridRel->projNodalData(vw,vw_h);
    
    for ( i=0; i < coarseNoNodes; i++ ) cov(i) = vw_h(i) - v_h(i)*w_h(i);

  } else if ( v->size() == gridRel->GetFineNoElms() ){

    if ( !gridRel->ElementRestrictionInitialized() ) gridRel->initNestElRestriction();

    vw.newsize(fineNoElms);
    cov.newsize(coarseNoElms);

    for ( i=0; i < fineNoElms; i++ ) vw(i) = (*v)(i) * (*w)(i);

    gridRel->projNestedElData(*v,v_h);
    gridRel->projNestedElData(*w,w_h);
    gridRel->projNestedElData(vw,vw_h);

    for ( i=0; i < coarseNoElms; i++ ) cov(i) = vw_h(i) - v_h(i)*w_h(i);

  } else{
    
    cout << "ERROR: Wrong size of vectors" << endl;
    exit(1);
    
  }

  return cov;
}

  

MV_Vector<real> ScaleExtrapolation::ExtrapolateCovariance( MV_Vector<real>* v, MV_Vector<real>* w, GridRelation* gridRelh22h, 
							   GridRelation* gridRelh24h, GridRelation* gridRel2h24h ){
  

  int fine_no_elms    = gridRelh24h->GetFineNoElms();
  int coarse_no_elms  = gridRelh24h->GetCoarseNoElms();
  int fine_no_nodes   = gridRelh24h->GetFineNoNodes();
  int coarse_no_nodes = gridRelh24h->GetCoarseNoNodes();

  int coarse_no_comp,fine_no_comp;
  if ( fine_no_nodes == v->size() ){
    coarse_no_comp = coarse_no_nodes;
    fine_no_comp   = fine_no_nodes;
  }
  if ( fine_no_elms  == v->size() ){
    coarse_no_comp = coarse_no_elms;
    fine_no_comp   = fine_no_elms;
  }
  
  MV_Vector<real> E2h = ComputeCovariance( v, w, gridRelh22h );
  MV_Vector<real> E4h = ComputeCovariance( v, w, gridRelh24h );

  /*
    MV_Vector<real> E2h_2h;
    if ( fine_no_nodes == v->size() ) gridRelh22h->projNodalData( E2h, E2h_2h );
    if ( fine_no_elms  == v->size() ) gridRelh22h->projNestedElData( E2h, E2h_2h );
  */

  if ( fine_no_nodes == v->size() ) gridRel2h24h->projNodalData( E2h );
  if ( fine_no_elms  == v->size() ) gridRel2h24h->projNestedElData( E2h );

  real mean_E2h = 0.0;
  real mean_E4h = 0.0;
  int i;
  for ( i=0; i < coarse_no_comp; i++ ){
    mean_E2h += E2h(i); 
    mean_E4h += E4h(i); 
  }
  mean_E2h /= coarse_no_comp;
  mean_E4h /= coarse_no_comp;
  
  real k = 3.0;
  real mean_qt = (mean_E4h-mean_E2h)/mean_E2h;
  real mean_wt = 1.0 - pow( mean_qt, -k );
  real mean_Eh = mean_wt * mean_E2h/( mean_qt - 1.0 );

  real var = 2.0;
  real cutoff_lower = 1.0 + (mean_qt-1.0)/var;
  real cutoff_upper = 1.0 + (mean_qt-1.0)*var;

  real wt;
  MV_Vector<real> Eh(fine_no_comp); 
  
  gridRelh24h->projNestedElData( E2h );
  gridRelh24h->projNestedElData( E4h );

  int corr = 0;
  for ( i=0; i < fine_no_comp; i++ ){
    wt = 1.0 - pow( (E4h(i)-E2h(i))/E2h(i) , -k );
    Eh(i) = wt * E2h(i)/((E4h(i)-E2h(i))/E2h(i) - 1.0);
    if ( (E4h(i)-E2h(i))/E2h(i) > cutoff_upper || (E4h(i)-E2h(i))/E2h(i) < cutoff_lower ){ 
      Eh(i) = mean_Eh;
      //Eh(i) = E2h_2h(i); 
      corr++;
    }
    if ( E2h(i) == 0.0 || E4h(i) == 0.0 ) Eh(i) = 0.0;
    //Eh(i) = E2h_2h(i); 
  }
  
  cout << "Corrected " << (100.0*corr)/(1.0*fine_no_comp) << " % of the dof." << endl;
  
  return Eh;
  
}


int ScaleExtrapolation::EstimateReynoldsStresses( MV_ColMat<real>* u, MV_ColMat<real>* tau_h, GridRelation* gridRelh22h, 
						  GridRelation* gridRelh24h, GridRelation* gridRel2h24h ){
  
  
  // This function estimates the Reynolds stresses tau_h on scale h of the velocity u. 
  // The estimation is based on extrapolation from coarser scales 2h and 4h, from 
  // the Ansatz tau_h = C(x)*h^(mu(x))

  // FIXME: Should be in Globals?
  bool element_based_extrapolation = true; 

  int no_comp;
  if ( element_based_extrapolation ){
    no_comp = gridRelh22h->GetFineNoElms();
  } else{
    no_comp = gridRelh22h->GetFineNoNodes();
  }
   
  tau_h->newsize(no_comp,6);

  MV_Vector<real> v(no_comp);
  MV_Vector<real> w(no_comp);
  
  MV_Vector<real> Eh; 

  int cnt = 0;

  Grid* fineGrid = gridRelh22h->GetFineGrid();
  int noNodesInElm = fineGrid->GetCell(0)->GetSize();
  if ( !(fineGrid->AllCellsEqual()) ){
    display->Error("Not implemented for hybrid grids.");
    exit(1);
  }
  
  int nsd = 3;

  int i,j,k,m,el,nod;
  for ( i=0; i < nsd; i++ ){  
    for ( j=i; j < nsd; j++ ){  

      if ( element_based_extrapolation ){
	for ( el=0; el < no_comp; el++ ){	
	  v(el) = 0.0;
	  for ( k=0; k < noNodesInElm; k++ ) v(el) += (*u)(fineGrid->GetCell(el)->GetNode(k),i);
	  v(el) /= noNodesInElm;

	  w(el) = 0.0;
	  for ( k=0; k < noNodesInElm; k++ ) w(el) += (*u)(fineGrid->GetCell(el)->GetNode(k),j);
	  w(el) /= noNodesInElm;
	}
      } else{
	for ( nod=0; nod < no_comp; nod++ ){	
	  v(nod) = (*u)(nod,i);
	  w(nod) = (*u)(nod,j);
	}
      }

      Eh = ExtrapolateCovariance( &v, &w, gridRelh22h, gridRelh24h, gridRel2h24h );

      for ( m=0; m < no_comp; m++ ) (*tau_h)(m,cnt) = Eh(m);
      cnt++;
    }
  }

  return 0;
}   




int ScaleExtrapolation::EstimateNonDivergenceForm( MV_ColMat<real>* u, MV_ColMat<real>* tau_h, GridRelation* gridRelh22h, 
						   GridRelation* gridRelh24h, GridRelation* gridRel2h24h ){
  
  
  // This function estimates the divergence of the Reynolds stresses on scale h,  
  // denoted here tau_h, of the velocity u, that is the 9 terms u_j * du_i/dx_j. 
  // The estimation is based on extrapolation from coarser scales 2h and 4h, from 
  // the Ansatz tau_h = C(x)*h^(mu(x))

  int no_elms = gridRelh22h->GetFineNoElms();
   
  tau_h->newsize(no_elms,9);

  MV_Vector<real> v(no_elms);
  MV_Vector<real> w(no_elms);
  
  int nsd = 3;

  Grid* fineGrid = gridRelh22h->GetFineGrid();
  int noNodesInElm = fineGrid->GetCell(0)->GetSize();
  if ( !(fineGrid->AllCellsEqual()) ){
    display->Error("Not implemented for hybrid grids.");
    exit(1);
  }
  
  MV_Vector<real> eni(noNodesInElm);
  MV_Vector<real> enj(noNodesInElm);
  MV_Vector<real> egrad(nsd);

  MV_Vector<real> Eh; 

  int cnt = 0;

  int i,j,k,m,el,nod;
  for ( i=0; i < nsd; i++ ){  
    for ( j=0; j < nsd; j++ ){  
      for ( el=0; el < no_elms; el++ ){

	for ( k=0; k < noNodesInElm; k++ ) eni(k) = (*u)(fineGrid->GetCell(el)->GetNode(k),i);
	for ( k=0; k < noNodesInElm; k++ ) enj(k) = (*u)(fineGrid->GetCell(el)->GetNode(k),j);

	v(el)  = ( enj(0) + enj(1) + enj(2) + enj(3) ) / noNodesInElm;

	// FIXME: Substitute this function
	//FE.evalGradientLoc(eni,egrad);
	w(el) = egrad(j);
      }

      Eh = ExtrapolateCovariance( &v, &w, gridRelh22h, gridRelh24h, gridRel2h24h );
      
      for ( m=0; m < no_elms; m++ ) (*tau_h)(m,cnt) = Eh(m);
      cnt++;
    }
  }
  
  return 0;
}   




int ScaleExtrapolation::ComputeReynoldsStresses( MV_ColMat<real>* u, MV_ColMat<real>* tau, GridRelation* gridRelh22h, 
						 GridRelation* gridRelh24h, GridRelation* gridRelh28h, 
						 GridRelation* gridRel2h28h, GridRelation* gridRel4h28h ){
  
  
  // This function computes the Reynolds stresses on scales 2h, 4h and 8h of a velocity u on scale h. 

  // FIXME: Should be in Globals?
  bool element_based_computation = true; 

  int no_comp;
  if ( element_based_computation ){
    no_comp = gridRelh22h->GetFineNoElms();
  } else{
    no_comp = gridRelh22h->GetFineNoNodes();
  }
   
  tau->newsize(no_comp,3*6);

  MV_Vector<real> v(no_comp);
  MV_Vector<real> w(no_comp);
  
  MV_Vector<real> cov; 

  int cnt = 0;

  int nsd = 3;

  Grid* fineGrid = gridRelh22h->GetFineGrid();
  int noNodesInElm = fineGrid->GetCell(0)->GetSize();
  if ( !(fineGrid->AllCellsEqual()) ){
    display->Error("Not implemented for hybrid grids.");
    exit(1);
  }

  int i,j,k,m,el,nod;
  for ( i=0; i < nsd; i++ ){  
    for ( j=i; j < nsd; j++ ){  

      if ( element_based_computation ){
	for ( el=0; el < no_comp; el++ ){	
	  v(el) = 0.0;
	  for ( k=0; k < noNodesInElm; k++ ) v(el) += (*u)(fineGrid->GetCell(el)->GetNode(k),i);
	  v(el) /= noNodesInElm;

	  w(el) = 0.0;
	  for ( k=0; k < noNodesInElm; k++ ) w(el) += (*u)(fineGrid->GetCell(el)->GetNode(k),j);
	  w(el) /= noNodesInElm;
	}
      } else{
	for ( nod=0; nod < no_comp; nod++ ){	
	  v(nod) = (*u)(nod,i);
	  w(nod) = (*u)(nod,j);
	}
      }

      cov = ComputeCovariance( &v, &w, gridRelh22h );
      for ( m=0; m < no_comp; m++ ) (*tau)(m,cnt)    = cov(m);
      cov = ComputeCovariance( &v, &w, gridRelh24h );
      for ( m=0; m < no_comp; m++ ) (*tau)(m,6+cnt)  = cov(m);
      cov = ComputeCovariance( &v, &w, gridRelh28h );
      for ( m=0; m < no_comp; m++ ) (*tau)(m,12+cnt) = cov(m);
      cnt++;
    }
  }

  return 0;
}   





int ScaleExtrapolation::ComputeNonDivergenceForm( MV_ColMat<real>* u, MV_ColMat<real>* tau, GridRelation* gridRelh22h, 
						  GridRelation* gridRelh24h, GridRelation* gridRelh28h, 
						  GridRelation* gridRel2h28h, GridRelation* gridRel4h28h ){
  
  
  // This function computes the divergence of the Reynolds stresses on scales 2h,4h, and 8h,   
  // of the velocity u on scale h, that is the 9 terms u_j * du_i/dx_j. 

  int no_comp;
  no_comp = gridRelh22h->GetFineNoElms();
   
  tau->newsize(no_comp,3*9);

  MV_Vector<real> v(no_comp);
  MV_Vector<real> w(no_comp);
  
  MV_Vector<real> cov; 

  int nsd = 3;

  Grid* fineGrid = gridRelh22h->GetFineGrid();
  int noNodesInElm = fineGrid->GetCell(0)->GetSize();
  if ( !(fineGrid->AllCellsEqual()) ){
    display->Error("Not implemented for hybrid grids.");
    exit(1);
  }

  MV_Vector<real> eni(noNodesInElm);
  MV_Vector<real> enj(noNodesInElm);
  MV_Vector<real> egrad(nsd);

  int cnt = 0;

  int i,j,k,m,el,nod;
  for ( i=0; i < nsd; i++ ){  
    for ( j=0; j < nsd; j++ ){  

      for ( el=0; el < no_comp; el++ ){

	for ( k=0; k < noNodesInElm; k++ ) eni(k) = (*u)(fineGrid->GetCell(el)->GetNode(k),i);
	for ( k=0; k < noNodesInElm; k++ ) enj(k) = (*u)(fineGrid->GetCell(el)->GetNode(k),j);

	v(el)  = ( enj(0) + enj(1) + enj(2) + enj(3) ) / noNodesInElm;

	// FIXME: Substitute this function
	//FE.evalGradientLoc(eni,egrad);
	w(el) = egrad(j);
      }

      cov = ComputeCovariance( &v, &w, gridRelh22h );
      for ( m=0; m < no_comp; m++ ) (*tau)(m,cnt)    = cov(m);
      cov = ComputeCovariance( &v, &w, gridRelh24h );
      for ( m=0; m < no_comp; m++ ) (*tau)(m,9+cnt)  = cov(m);
      cov = ComputeCovariance( &v, &w, gridRelh28h );
      for ( m=0; m < no_comp; m++ ) (*tau)(m,18+cnt) = cov(m);
      cnt++;
    }
  }

  return 0;
}   





