// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "TranProblem.hh"

//-----------------------------------------------------------------------------
TranProblem::TranProblem(Grid *grid,
								 MV_Vector < GridRelation > *GR)
{
  // Initialize the grid
  this->grid = grid;
  
  // Initialize  GRIDREL
  GRIDREL = GR;
  
  // Update internal variables
  Update();
  
  // Open the OpenDX file
  opendx_residuals = new OpenDX("residuals.dx", 3, 3, 1, 1);
  opendx_residuals->SetLabel(0, "R1");
  opendx_residuals->SetLabel(1, "R2");
  opendx_residuals->SetLabel(2, "R4");
  
  // Initialize internal constants
  pi = KW_PI;
}
//-----------------------------------------------------------------------------
TranProblem::~TranProblem()
{
	// Delete the OpenDX file
	delete opendx_residuals;
}
//-----------------------------------------------------------------------------
Grid *TranProblem::GetGrid()
{
	return grid;
}
//-----------------------------------------------------------------------------
int TranProblem::CheckOutputLevel()
{
	// Check the output level
	if (printGrid > compGrid) {
		display->Warning("Output level cannot be greater than highest level. Setting output level = %i (=highest level)",
							  compGrid);
		globals->Set("current output level", compGrid);
	}

	return 0;
}
//-----------------------------------------------------------------------------
int TranProblem::Update()
{
	// Update output settings
	settings->Get("write residuals", &write_residuals);
	settings->Get("write data", &write_data);

	// Update what grids to use to compute and write data 
	globals->Get("current computational level", &compGrid);
	globals->Get("current output level", &printGrid);

	// Update internal variables 

	//no_nodes = (*GRID) (compGrid).GetNoNodes();
	//no_elms = (*GRID) (compGrid).GetNoCells();
	no_nodes = grid->GetNoNodes();
	no_elms  = grid->GetNoCells();

	settings->Get("dimension", &nsd);

	//printNoNodes = (*GRID) (printGrid).GetNoNodes();
	//printNoElms  = (*GRID) (printGrid).GetNoCells();

	printNoNodes = grid->GetNoNodes();
	printNoElms  = grid->GetNoCells();

	settings->Get("start time", &T0);
	settings->Get("final time", &T);

	// Update smallest element diameter 
	GetMeshSize(grid_h);

	// Set output frame to zero 
	current_frame = 0;

	return 0;
}
//-----------------------------------------------------------------------------
bool TranProblem::SaveSample()
{
	// Check if we should save a sample

	bool save_sample = false;

	int ops;
	settings->Get("output samples", &ops);
	real M = real(ops);

	real n = current_frame;
	real sample_time;

	if (dt < 0) {
		sample_time = T - (T - T0) / M * n;
		if (t <= (sample_time + KW_EPS))
			save_sample = true;
	} else {
		sample_time = T0 + (T - T0) / M * n;
		if (t >= (sample_time - KW_EPS))
			save_sample = true;
	}

	if (save_sample)
		current_frame += 1;

	return (save_sample);
}
//-----------------------------------------------------------------------------
int TranProblem::GetContResidual(MV_Vector < real > &res)
{
	res = contResidual;

	return 0;
}
//-----------------------------------------------------------------------------
int TranProblem::AttachGrid(Grid *grid)
{
  this->grid = grid;

  return 0;
}
//-----------------------------------------------------------------------------
int TranProblem::AttachGridRelation(MV_Vector < GridRelation > *GR)
{
	GRIDREL = GR;

	return 0;
}
//-----------------------------------------------------------------------------
int TranProblem::AttachField(MV_Vector < Field > *F)
{
	FIELD = F;

	return 0;
}
//-----------------------------------------------------------------------------
real TranProblem::LocalMeshSize(int el)
{
  // Compute element diameter
  
  real gh = 2.0 * grid->GetCell(el)->ComputeCircumRadius(grid);

  //display->Message(0,"grid size = %f",gh);
  
  return gh;
}
//-----------------------------------------------------------------------------
int TranProblem::GetMeshSize(real & gh)
{
	// Compute smallest element diameter in grid

	gh = 1000000.0;

	real lms;

	int el;
	for (el = 0; el < no_elms; el++) {
		lms = LocalMeshSize(el);
		if (lms < gh)
			gh = lms;
	}

	return 0;
}
//-----------------------------------------------------------------------------
bool TranProblem::CheckIfStationary()
{
	// Check if the solution is stationary 

	bool stationary_solution = false;

	real maxnorm_du = 0.0;
	int i, j;
	int noeq = uk.size(1);
	for (i = 0; i < no_nodes; i++) {
		for (j = 0; j < noeq; j++) {
			if (fabs(uk(i, j) - uk_prev_TS(i, j)) > maxnorm_du)
				maxnorm_du =
				    fabs(uk(i, j) - uk_prev_TS(i, j));
		}
	}
	if (maxnorm_du / dt < stat_tol) {
		cout << "Stationary limit reached, at time t = " << t <<
		    ", max du/dt = " << maxnorm_du /
		    dt << " (statinary tolerance = " << stat_tol << ")" <<
		    endl;
		stationary_solution = true;
	}

	return 0;
}
//-----------------------------------------------------------------------------
