// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __TRAN_PROBLEM_HH
#define __TRAN_PROBLEM_HH

#include "Globals.hh"
#include "Settings.hh"
#include "kw_constants.h"
#include "Field.hh"
#include "GridRelation.h"
#include "OpenDX.hh"
#include "Matlab.hh"

extern Settings *settings;
extern Globals *globals;

///
class TranProblem{
public:
  
  /// Constructors
  TranProblem() {};
  TranProblem(Grid *grid, MV_Vector<GridRelation>* GR);
  
  /// Destructor
  ~TranProblem();

  /// Solve transient problem
  virtual  
  int Solve() = 0;
  
  /// Returns a pointer to the Grid vector associated to TranProblem
  Grid *GetGrid();

  /// Attach pointer to Grid vector
  int AttachGrid(Grid *grid);

  /// Attach pointer to GridRelation vector
  int AttachGridRelation(MV_Vector<GridRelation>* GR);

  /// Attach pointer to Field vector
  int AttachField( MV_Vector<Field>* F );

  /// Return continuous residual
  int GetContResidual( MV_Vector<real>& res );

 protected:

  /// Updates internal variables 
  int Update();
  
  /// set initial data 
  virtual  
  int SetInitialData() = 0;
  
  /// set internal tolerances for iterative method 
  virtual
  int SetTolerances() = 0;

  /// computes continuous residuals
  virtual
  int CompContResidual( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt, MV_Vector<real>& contResidual ) = 0;

  /// Writes different norms of the solution to file
  virtual 
  int WriteData( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt ) = 0;

  /// return global resp. local mesh size
  int GetMeshSize( real& gh );
  real LocalMeshSize( int el );

  /// Checks if the solution has reached a steady state
  bool CheckIfStationary();

  /// checks output level wrt computational level
  int CheckOutputLevel();

  /// check if to save data or not  
  bool SaveSample();
  int  current_frame;

  /// Matlab object for saving data
  Matlab *matlab_data;

  /// OpenDX object for writing residuals to file
  OpenDX *opendx_residuals; 

  /// A pointer to the grid
  Grid *grid;
  
  /// internal pointers to vectors containing Grid, GridRelation, and Field
  MV_Vector<GridRelation>* GRIDREL;
  MV_Vector<Field>* FIELD;

  /// uk solution on current time step, uk_pres_TS solution on previous time step
  MV_ColMat<real> uk;
  MV_ColMat<real> uk_prev_TS;

  /// internal variables
  int printGrid;
  int compGrid;

  int nsd;  
  int no_elms;
  int no_nodes;

  int printNoNodes;
  int printNoElms;
  
  real dt,t,T0,T;
  real grid_h;

  real stat_tol;

  real pi;

  MV_Vector<real> contResidual;
  
  bool write_data;
  bool write_residuals;

};

#endif
